import logging
from uuid import uuid4 as new_uuid

import numpy as np
import pandas as pd
import pytest

import btrdb.utils.timez
from btrdb.utils.timez import currently_as_ns, ns_delta

try:
    import pyarrow as pa
except ImportError:
    pa = None


def test_stream_methods(conn, tmp_collection):
    s1 = conn.create(
        new_uuid(),
        tmp_collection,
        tags={"name": "s1", "unit": "foo"},
        annotations={"test": "bar"},
    )


def test_grpc_insert_and_values(conn, tmp_collection):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    data = []
    duration = 100
    for i in range(duration):
        data.append([t + i, i])
    s.insert(data)
    fetched_data = s.values(start=t, end=t + duration)
    assert len(fetched_data) == len(data)
    for i, (p, _) in enumerate(fetched_data):
        assert p.time == data[i][0]
        assert p.value == data[i][1]


def test_arrow_insert_and_values(
    conn, tmp_collection, single_stream_values_arrow_schema
):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = 100
    for i in range(duration):
        times.append(t + i)
        values.append(i)
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    fetched_data = s.arrow_values(start=t, end=t + duration)
    assert data == fetched_data


def test_arrow_values_template_schema(conn, tmp_collection):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = [100, 200, 300, 400]
    values = [1.0, 2.0, 3.0, 4.0]
    s.insert(list(zip(times, values)))
    schema = pa.schema(
        [
            pa.field("t", pa.int64(), nullable=False),
            pa.field("v", pa.float32(), nullable=False),
        ]
    )
    expected = pa.Table.from_arrays([pa.array(times), pa.array(values)], schema=schema)
    fetched_data = s.arrow_values(start=times[0], end=times[-1] + 1, schema=schema)
    assert expected == fetched_data
    assert expected.schema.equals(fetched_data.schema)


@pytest.mark.parametrize(
    "merge_policy,duplicates_expected",
    [("never", True), ("equal", True), ("retain", False), ("replace", False)],
)
def test_arrow_insert_duplicate_timestamps(
    conn,
    tmp_collection,
    single_stream_values_arrow_schema,
    merge_policy,
    duplicates_expected,
):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    t1 = [100, 105, 110, 110, 120]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    s1.insert(list(zip(t1, d1)), merge=merge_policy)
    df = s1.arrow_values(start=100, end=121).to_pandas().set_index("time")
    assert any(df.index.duplicated().tolist()) == duplicates_expected


def test_arrow_values_table_schema(
    conn, tmp_collection, single_stream_values_arrow_schema
):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = 100
    for i in range(duration):
        times.append(t + i)
        values.append(i)
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    fetched_data = s.arrow_values(start=t, end=t + duration)
    assert single_stream_values_arrow_schema.equals(fetched_data.schema)


def test_arrow_values_table_schema_2(
    conn, tmp_collection, single_stream_values_arrow_schema
):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = 100
    for i in range(duration):
        times.append(t + i)
        values.append(i)
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    fetched_data = s.arrow_values(start=t, end=t + duration)
    assert single_stream_values_arrow_schema.equals(fetched_data.schema)


def test_arrow_windows(
    conn,
    tmp_collection,
    single_stream_windows_all_stats_arrow_schema,
    single_stream_values_arrow_schema,
):
    rng = np.random.default_rng(seed=7)
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = btrdb.utils.timez.ns_delta(minutes=1)
    nvals = 100
    for i in range(nvals):
        times.append(t + i * duration)
        values.append(i + rng.random())
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    # group these 5 at a time to go with the 5 min windows from below
    mean_dat = np.mean(np.asarray(values).reshape(-1, 5), axis=1)
    min_dat = np.min(np.asarray(values).reshape(-1, 5), axis=1)
    max_dat = np.max(np.asarray(values).reshape(-1, 5), axis=1)
    count_dat = [np.asarray(values).reshape(-1, 5).shape[1]] * int(
        np.asarray(values).shape[0] / 5
    )
    stddev_dat = np.std(np.asarray(values).reshape(-1, 5), axis=1)
    time_dat = [t for i, t in enumerate(times) if i % 5 == 0]
    window_table = pa.Table.from_arrays(
        [time_dat, min_dat, mean_dat, max_dat, count_dat, stddev_dat],
        schema=single_stream_windows_all_stats_arrow_schema,
    )
    s.arrow_insert(data)
    # 5 min windows
    fetched_data = s.arrow_windows(
        start=t, end=t + duration * nvals + 1, width=duration * 5
    )
    fetched_df = fetched_data.to_pandas()
    fetched_df["time"] = fetched_df["time"].astype(int)
    window_df = window_table.to_pandas()
    window_df["time"] = window_df["time"].astype(int)
    assert np.allclose(fetched_df.values, window_df.values, rtol=0, atol=1e-9)


@pytest.mark.xfail
def test_aligned_windows_no_flush(
    conn,
    tmp_collection,
    single_stream_windows_all_stats_arrow_schema,
    single_stream_values_arrow_schema,
):
    rng = np.random.default_rng(seed=7)
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = btrdb.utils.timez.ns_delta(minutes=1)
    nvals = 100
    for i in range(nvals):
        times.append(t + i * duration)
        values.append(i + rng.random())
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    fetched_data = s.arrow_aligned_windows(
        start=t,
        end=t + duration * nvals,
        pointwidth=btrdb.utils.general.pointwidth.from_nanoseconds(duration * 6),
    )
    fetched_df = fetched_data.to_pandas()
    assert fetched_df["stddev"].isna().sum() == 0


def test_arrow_aligned_windows_vs_aligned_windows(
    conn,
    tmp_collection,
    single_stream_windows_all_stats_arrow_schema,
    single_stream_values_arrow_schema,
):
    rng = np.random.default_rng(seed=7)
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = btrdb.utils.timez.ns_delta(minutes=1)
    nvals = 100
    for i in range(nvals):
        times.append(t + i * duration)
        values.append(i + rng.random())
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    s.flush()
    fetched_data = s.arrow_aligned_windows(
        start=t,
        end=t + duration * nvals,
        pointwidth=btrdb.utils.general.pointwidth.from_nanoseconds(duration * 6),
    )
    fetched_df = fetched_data.to_pandas()
    fetched_df["time"] = fetched_df["time"].astype(int)
    fetched_df["count"] = fetched_df["count"].astype(int)

    other_method = s.aligned_windows(
        start=t,
        end=t + duration * nvals,
        pointwidth=btrdb.utils.general.pointwidth.from_nanoseconds(duration * 6),
    )
    other_method_data = []
    for statpt, ver in other_method:
        tmp = dict()
        tmp["time"] = statpt.time
        tmp["min"] = statpt.min
        tmp["mean"] = statpt.mean
        tmp["max"] = statpt.max
        tmp["count"] = statpt.count
        tmp["stddev"] = statpt.stddev
        other_method_data.append(tmp)
    other_method_df = pd.DataFrame(other_method_data)
    assert fetched_data.schema.equals(single_stream_windows_all_stats_arrow_schema)
    assert fetched_df.equals(other_method_df)


def compare_arrow_windows_vs_windows(
    conn,
    tmp_collection,
    single_stream_windows_all_stats_arrow_schema,
    single_stream_values_arrow_schema,
):
    rng = np.random.default_rng(seed=7)
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    t = currently_as_ns()
    times = []
    values = []
    duration = btrdb.utils.timez.ns_delta(minutes=1)
    nvals = 100
    for i in range(nvals):
        times.append(t + i * duration)
        values.append(i + rng.random())
    data = pa.Table.from_arrays(
        [
            pa.array(times),
            pa.array(values),
        ],
        schema=single_stream_values_arrow_schema,
    )
    s.arrow_insert(data)
    fetched_data = s.arrow_windows(
        start=t,
        end=t + duration * nvals,
        width=duration * 5,
    )
    fetched_df = fetched_data.to_pandas()
    fetched_df["time"] = fetched_df["time"].astype(int)
    fetched_df["count"] = fetched_df["count"].astype(int)

    other_method = s.windows(start=t, end=t + duration * nvals, width=duration * 5)
    other_method_data = []
    for statpt, ver in other_method:
        tmp = dict()
        tmp["time"] = statpt.time
        tmp["min"] = statpt.min
        tmp["mean"] = statpt.mean
        tmp["max"] = statpt.max
        tmp["count"] = statpt.count
        tmp["stddev"] = statpt.stddev
        other_method_data.append(tmp)
    other_method_df = pd.DataFrame(other_method_data)
    assert fetched_data.schema.equals(single_stream_windows_all_stats_arrow_schema)
    assert fetched_df.equals(other_method_df)


def test_arrow_empty_values(conn, tmp_collection):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    data = s.arrow_values(start=100, end=1000)
    assert len(data["time"]) == 0
    assert len(data["value"]) == 0


def test_arrow_empty_values_schema(conn, tmp_collection):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    data = s.arrow_values(start=100, end=1000)
    schema = pa.schema(
        [
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field("value", pa.float64(), nullable=False),
        ]
    )
    assert schema.equals(data.schema)


def test_stream_annotation_update(conn, tmp_collection):
    s = conn.create(
        new_uuid(),
        tmp_collection + "foo/",
        tags={"name": "s"},
        annotations={"foo": "bar"},
    )
    annotations1, version1 = s.annotations()
    assert version1 == 1
    assert annotations1["foo"] == "bar"
    s.update(annotations={"foo": "baz"})
    annotations2, version2 = s.annotations()
    assert version2 > version1
    assert annotations2["foo"] == "baz"
    s.update(annotations={}, replace=True)
    annotations3, version3 = s.annotations()
    assert len(annotations3) == 0
    assert version3 > version2



# tests.test_timez
# Testing for the btrdb.timez module
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: test_timez.py [] allen@pingthings.io $

"""
Testing for the btrdb.timez module
"""

##########################################################################
## Imports
##########################################################################

import datetime

import numpy as np
import pytest
import pytz
from freezegun import freeze_time

from btrdb.utils.timez import (
    currently_as_ns,
    datetime_to_ns,
    ns_delta,
    ns_to_datetime,
    to_nanoseconds,
)

##########################################################################
## Initialization Tests
##########################################################################


class TestCurrentlyAsNs(object):
    def test_currently_as_ns(self):
        """
        Assert currently_as_ns returns correct value
        """
        expected = 1514808000000000000
        with freeze_time("2018-01-01 12:00:00 -0000"):
            assert currently_as_ns() == expected


class TestDatetimeToNs(object):
    def test_datetime_to_ns_naive(self):
        """
        Assert datetime_to_ns handles naive datetime
        """
        dt = datetime.datetime(2018, 1, 1, 12)
        localized = pytz.utc.localize(dt)
        expected = int(localized.timestamp() * 1e9)

        assert dt.tzinfo is None
        assert datetime_to_ns(dt) == expected

    def test_datetime_to_ns_aware(self):
        """
        Assert datetime_to_ns handles tz aware datetime
        """
        eastern = pytz.timezone("US/Eastern")
        dt = datetime.datetime(2018, 1, 1, 17, tzinfo=eastern)
        expected = int(dt.astimezone(pytz.utc).timestamp() * 1e9)

        assert dt.tzinfo is not None
        assert datetime_to_ns(dt) == expected


class TestNsToDatetime(object):
    def test_ns_to_datetime_is_utc(self):
        """
        Assert ns_to_datetime returns UTC aware datetime
        """
        dt = datetime.datetime.utcnow()
        ns = int(dt.timestamp() * 1e9)

        assert dt.tzinfo is None
        assert ns_to_datetime(ns).tzinfo == pytz.UTC

    def test_ns_to_datetime_is_correct(self):
        """
        Assert ns_to_datetime returns correct datetime
        """
        val = 1514808000000000000
        expected = datetime.datetime(2018, 1, 1, 12, tzinfo=pytz.UTC)

        assert ns_to_datetime(val) == expected


class TestToNanoseconds(object):
    def test_datetime_to_ns_naive(self):
        """
        Assert to_nanoseconds handles naive datetime
        """
        dt = datetime.datetime(2018, 1, 1, 12)
        localized = pytz.utc.localize(dt)
        expected = int(localized.timestamp() * 1e9)

        assert dt.tzinfo is None
        assert to_nanoseconds(dt) == expected

    def test_datetime_to_ns_aware(self):
        """
        Assert to_nanoseconds handles tz aware datetime
        """
        eastern = pytz.timezone("US/Eastern")
        dt = datetime.datetime(2018, 1, 1, 17, tzinfo=eastern)
        expected = int(dt.astimezone(pytz.utc).timestamp() * 1e9)

        assert dt.tzinfo is not None
        assert to_nanoseconds(dt) == expected

    def test_str(self):
        """
        Assert to_nanoseconds handles RFC3339 format
        """
        dt = datetime.datetime(2018, 1, 1, 12, tzinfo=pytz.utc)
        expected = int(dt.timestamp() * 1e9)

        dt_str = "2018-1-1 12:00:00.0-0000"
        assert dt.tzinfo is not None
        assert to_nanoseconds(dt_str) == expected

        dt_str = "2018-1-1 7:00:00.0-0500"
        dt = datetime.datetime(2018, 1, 1, 12, tzinfo=pytz.timezone("US/Eastern"))
        assert dt.tzinfo is not None
        assert to_nanoseconds(dt_str) == expected

        dt_str = "2018-01-15 07:32:49"
        dt = datetime.datetime(2018, 1, 15, 7, 32, 49, tzinfo=pytz.utc)
        expected = int(dt.timestamp() * 1e9)
        assert to_nanoseconds(dt_str) == expected

    def test_str_midnight(self):
        """
        Test parse a date at midnight
        """
        expected = datetime.datetime(2019, 4, 7, tzinfo=pytz.utc)
        expected = int(expected.timestamp() * 1e9)
        assert to_nanoseconds("2019-04-07") == expected

    def test_str_raise_valueerror(self):
        """
        Assert to_nanoseconds raises on invalid str
        """
        dt_str = "01 Jan 2018 12:00:00 -0000"
        with pytest.raises(ValueError, match="RFC3339") as exc:
            to_nanoseconds(dt_str)

    def test_int(self):
        """
        Assert to_nanoseconds handles int
        """
        assert 42 == to_nanoseconds(42)

    def test_float(self):
        """
        Assert to_nanoseconds handles float
        """
        assert 42 == to_nanoseconds(42.0)

    def test_float_raise_valueerror(self):
        """
        Assert to_nanoseconds raises on invalid float
        """
        with pytest.raises(ValueError) as exc:
            to_nanoseconds(42.5)
        assert "can only convert whole numbers" in str(exc)

    def test_float(self):
        """
        Assert to_nanoseconds handles float
        """
        dt = datetime.datetime(2018, 1, 1, 12, tzinfo=pytz.utc)
        expected = int(dt.timestamp() * 1e9)

        dt64 = np.datetime64("2018-01-01T12:00")
        assert expected == to_nanoseconds(dt64)


class TestToNsDelta(object):
    def test_ns_delta(self):
        """
        Assert ns_delta converts inputs properly
        """
        val = ns_delta(1, 2, 1, 3, 1, 23, 1)
        assert val == 93663001023001

    def test_ns_delta_precision(self):
        """
        Assert ns_delta deals with real inputs
        """
        val = ns_delta(days=365, minutes=0.5, nanoseconds=1)
        assert val == int(1e9 * 60 * 60 * 24 * 365) + int(1e9 * 30) + 1

    def test_returns_int(self):
        """
        Assert ns_delta returns int if floats used for arguments
        """
        val = ns_delta(1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1)
        assert val == int(
            1 + 1e3 + 1e6 + 1e9 + (1e9 * 60) + (1e9 * 60 * 60) + (1e9 * 60 * 60 * 24)
        )
        assert isinstance(val, int)



# tests.test_stream
# Testing package for the btrdb stream module
#
# Author:   PingThings
# Created:  Wed Jan 02 19:26:20 2019 -0500
#
# For license information, see LICENSE.txt
# ID: test_stream.py [] allen@pingthings.io $

"""
Testing package for the btrdb stream module
"""
import concurrent.futures
import datetime
import json
import re
import sys
import uuid
from unittest.mock import Mock, PropertyMock, call, patch

import pytest
import pytz

from btrdb import MAXIMUM_TIME, MINIMUM_TIME
from btrdb.conn import BTrDB
from btrdb.endpoint import Endpoint
from btrdb.exceptions import (
    BTrDBError,
    InvalidCollection,
    InvalidOperation,
    NoSuchPoint,
    StreamNotFoundError,
)
from btrdb.grpcinterface import btrdb_pb2
from btrdb.point import RawPoint, StatPoint
from btrdb.stream import INSERT_BATCH_SIZE, Stream, StreamFilter, StreamSet

##########################################################################
## Imports
##########################################################################


RawPointProto = btrdb_pb2.RawPoint
StatPointProto = btrdb_pb2.StatPoint
EST = pytz.timezone("America/New_York")


##########################################################################
## Fixtures
##########################################################################


@pytest.fixture
def stream1():
    uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
    stream = Mock(Stream)
    stream.version = Mock(return_value=11)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=10, value=1), 11))
    type(stream).collection = PropertyMock(return_value="fruits/apple")
    type(stream).name = PropertyMock(return_value="gala")
    stream.tags = Mock(return_value={"name": "gala", "unit": "volts"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "red"}, 11))
    stream._btrdb = Mock()
    stream._btrdb._executor = concurrent.futures.ThreadPoolExecutor()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=False)
    return stream


@pytest.fixture
def stream2():
    uu = uuid.UUID("17dbe387-89ea-42b6-864b-f505cdb483f5")
    stream = Mock(Stream)
    stream.version = Mock(return_value=22)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=20, value=1), 22))
    type(stream).collection = PropertyMock(return_value="fruits/orange")
    type(stream).name = PropertyMock(return_value="blood")
    stream.tags = Mock(return_value={"name": "blood", "unit": "amps"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "orange"}, 22))
    stream._btrdb = Mock()
    stream._btrdb._executor = Mock()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=False)
    return stream


@pytest.fixture
def stream3():
    uu = uuid.UUID("17dbe387-89ea-42b6-864b-e2ef0d22a53b")
    stream = Mock(Stream)
    stream.version = Mock(return_value=33)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=30, value=1), 33))
    type(stream).collection = PropertyMock(return_value="fruits/banana")
    type(stream).name = PropertyMock(return_value="yellow")
    stream.tags = Mock(return_value={"name": "yellow", "unit": "watts"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "yellow"}, 3))
    stream._btrdb = Mock()
    stream._btrdb._executor = Mock()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=False)
    return stream


@pytest.fixture
def arrow_stream3():
    uu = uuid.UUID("17dbe387-89ea-42b6-864b-f505cdb483f5")
    stream = Mock(Stream)
    stream.version = Mock(return_value=33)
    stream.uuid = Mock(return_value=uu)
    stream.nearest = Mock(return_value=(RawPoint(time=20, value=1), 22))
    type(stream).collection = PropertyMock(return_value="fruits/orange")
    type(stream).name = PropertyMock(return_value="blood")
    stream.tags = Mock(return_value={"name": "blood", "unit": "amps"})
    stream.annotations = Mock(return_value=({"owner": "ABC", "color": "orange"}, 22))
    stream._btrdb = Mock()
    stream._btrdb._executor = Mock()
    stream._btrdb._ARROW_ENABLED = Mock(return_value=True)
    return stream


##########################################################################
## Stream Tests
##########################################################################


class TestStream(object):
    def test_create(self):
        """
        Assert we can create the object
        """
        Stream(None, "FAKE_UUID")

    def test_repr_str(self):
        """
        Assert the repr and str output are correct
        """
        COLLECTION = "relay/foo"
        NAME = "LINE222VA-ANG"
        stream = Stream(None, "FAKE")
        stream._collection = COLLECTION
        stream._tags = {"name": NAME}

        expected = "<Stream collection={} name={}>".format(COLLECTION, NAME)
        assert stream.__repr__() == expected
        assert stream.__str__() == expected

    def test_refresh_metadata(self):
        """
        Assert refresh_metadata calls Endpoint.streamInfo
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, {}, None))
        endpoint.info = Mock(return_value={"majorVersion": 5, "minorVersion": 30})
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        stream.refresh_metadata()
        stream._btrdb.ep.streamInfo.assert_called_once_with(uu, False, True)

    def test_refresh_metadata_deserializes_annotations(self):
        """
        Assert refresh_metadata deserializes annotation values
        """
        uu = uuid.uuid4()
        serialized = {
            "acronym": "VPHM",
            "description": "El Segundo PMU 42 Ean",
            "devacronym": "PMU!EL_SEG_PMU_42",
            "enabled": "true",
            "id": "76932ae4-09bc-472c-8dc6-64fea68d2797",
            "phase": "A",
            "label": "null",
            "frequency": "30",
            "control": "2019-11-07 13:21:23.000000-0500",
            "calibrate": '{"racf": 1.8, "pacf": 0.005}',
        }
        expected = {
            "acronym": "VPHM",
            "description": "El Segundo PMU 42 Ean",
            "devacronym": "PMU!EL_SEG_PMU_42",
            "enabled": True,
            "id": "76932ae4-09bc-472c-8dc6-64fea68d2797",
            "phase": "A",
            "label": None,
            "frequency": 30,
            "control": "2019-11-07 13:21:23.000000-0500",
            "calibrate": {"racf": 1.8, "pacf": 0.005},
        }

        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, serialized, None))
        endpoint.info = Mock(return_value={"majorVersion": 5, "minorVersion": 30})
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        stream.refresh_metadata()
        assert stream.annotations()[0] == expected

    def test_stream_name_property(self):
        """
        Assert name property comes from tags
        """
        name = "LINE222VA-ANG"
        stream = Stream(None, "FAKE_UUID")
        stream._tags = {"name": name}
        assert stream.name == name

    def test_stream_unit_property(self):
        """
        Assert unit property comes from tags
        """
        unit = "whales"
        stream = Stream(None, "FAKE_UUID")
        stream._tags = {"unit": unit}
        assert stream.unit == unit

    ##########################################################################
    ## update tests
    ##########################################################################

    def test_update_arguments(self):
        """
        Assert update raises errors on invalid arguments
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.info = Mock(return_value={"majorVersion": 5, "minorVersion": 30})
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        # no arguments
        with pytest.raises(ValueError) as exc:
            stream.update()
        assert "must supply" in str(exc)

        # tags not dict
        with pytest.raises(TypeError) as exc:
            stream.update(tags=[])
        assert "tags must be of type dict" in str(exc)

        # annotations not dict
        with pytest.raises(TypeError) as exc:
            stream.update(annotations=[])
        assert "annotations must be of type dict" in str(exc)

        # collection not string
        with pytest.raises(InvalidCollection) as exc:
            stream.update(collection=42)
        assert "collection must be of type string" in str(exc)

    def test_update_tags(self):
        """
        Assert update calls correct Endpoint methods for tags update
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.info = Mock(return_value={"majorVersion": 5, "minorVersion": 30})
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, {}, None))
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        tags = {"cat": "dog"}

        stream.update(tags=tags)
        stream._btrdb.ep.setStreamTags.assert_called_once_with(
            uu=uu, expected=42, tags=tags, collection="koala"
        )
        stream._btrdb.ep.setStreamAnnotations.assert_not_called()

    def test_update_collection(self):
        """
        Assert update calls correct Endpoint methods for collection update
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, {}, None))
        endpoint.info = Mock(return_value={"majorVersion": 5, "minorVersion": 30})
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu, property_version=42)

        collection = "giraffe"

        stream.update(collection=collection)
        stream._btrdb.ep.setStreamTags.assert_called_once_with(
            uu=uu, expected=42, tags={}, collection=collection
        )

    def test_update_annotations(self):
        """
        Assert update calls correct Endpoint methods for annotations update
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, {}, None))
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        # Test realistic annotations with multiple types
        annotations = {
            "acronym": "VPHM",
            "description": "El Segundo PMU 42 Ean",
            "devacronym": "PMU!EL_SEG_PMU_42",
            "enabled": True,
            "id": uuid.UUID("76932ae4-09bc-472c-8dc6-64fea68d2797"),
            "phase": "A",
            "label": None,
            "frequency": 30,
            "control": EST.localize(datetime.datetime(2019, 11, 7, 13, 21, 23)),
            "calibrate": {"racf": 1.8, "pacf": 0.005},
        }

        stream.refresh_metadata()
        stream.update(annotations=annotations)
        stream._btrdb.ep.setStreamAnnotations.assert_called_once_with(
            uu=uu,
            expected=42,
            changes={
                "acronym": "VPHM",
                "description": "El Segundo PMU 42 Ean",
                "devacronym": "PMU!EL_SEG_PMU_42",
                "enabled": "true",
                "id": "76932ae4-09bc-472c-8dc6-64fea68d2797",
                "phase": "A",
                "label": "null",
                "frequency": "30",
                "control": "2019-11-07 13:21:23.000000-0500",
                "calibrate": '{"racf": 1.8, "pacf": 0.005}',
            },
            removals=[],
        )
        stream._btrdb.ep.setStreamTags.assert_not_called()

    def test_update_annotations_nested_conversions(self):
        """
        Assert update correctly encodes nested annotation data
        """
        annotations = {
            "num": 10,
            "float": 1.3,
            "string": "the quick brown fox is 10",
            "nested": {
                "num": 11,
                "float": 1.3,
                "string": "the quick brown fox is 11",
                "nested": {
                    "num": 12,
                    "float": 1.3,
                    "string": "the quick brown fox is 12",
                },
            },
        }
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(
            return_value=("koala", 42, {}, {"foo": "42 Cherry Hill"}, None)
        )
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        stream.refresh_metadata()
        stream.update(annotations=annotations)

        # spot check a nested value for Python 3.4 and 3.5 compatability
        changes = stream._btrdb.ep.setStreamAnnotations.call_args[1]["changes"]
        assert changes["nested"].__class__ == str
        assert json.loads(changes["nested"]) == annotations["nested"]

        # check all args if Python > 3.5
        if sys.version_info[0] > 3.5:
            stream._btrdb.ep.setStreamAnnotations.assert_called_once_with(
                uu=uu,
                expected=42,
                changes={
                    "num": "10",
                    "float": "1.3",
                    "string": '"the quick brown fox is 10"',
                    "nested": '{"num": 11, "float": 1.3, "string": "the quick brown fox is 11", "nested": {"num": 12, "float": 1.3, "string": "the quick brown fox is 12"}}',
                },
            )

    def test_update_annotations_no_encoder(self):
        """
        Assert update annotations works with None as encoder argument
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("koala", 42, {}, {}, None))
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        annotations = {"foo": "this is a string", "bar": "3.14"}

        stream.refresh_metadata()
        stream.update(annotations=annotations, encoder=None)
        stream._btrdb.ep.setStreamAnnotations.assert_called_once_with(
            uu=uu,
            expected=42,
            changes=annotations,
            removals=[],
        )

        # TODO: mock json.dumps
        # assert mock_dumps.assert_not_called()

    def test_update_annotations_replace(self):
        """
        Assert that replace argument will add proper keys to removals array in
        endpoint call.
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(
            return_value=("koala", 42, {}, {"phase": "A", "source": "PJM"}, None)
        )
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        annotations = {
            "foo": "this is a string",
            "phase": "A",
        }

        stream.refresh_metadata()

        # remove one of the keys and add new ones
        stream.update(annotations=annotations, replace=True)
        stream._btrdb.ep.setStreamAnnotations.assert_called_once_with(
            uu=uu,
            expected=42,
            changes=annotations,
            removals=["source"],
        )

        # clear annotations
        stream.update(annotations={}, replace=True)
        stream._btrdb.ep.setStreamAnnotations.assert_called_with(
            uu=uu,
            expected=42,
            changes={},
            removals=["phase", "source"],
        )

    ##########################################################################
    ## exists tests
    ##########################################################################

    def test_exists_cached_value(self):
        """
        Assert exists first uses cached value
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu, known_to_exist=True)
        stream.refresh_metadata = Mock()

        assert stream.exists()
        stream.refresh_metadata.assert_not_called()

    def test_exists(self):
        """
        Assert exists refreshes data if value is unknown
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)
        stream.refresh_metadata = Mock(return_value=True)

        assert stream.exists()
        assert stream.refresh_metadata.call_count == 1

    def test_exists_returns_false_on_404(self):
        """
        Assert exists returns False on 404 error
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)
        stream.refresh_metadata = Mock(
            side_effect=StreamNotFoundError("stream not found with provided uuid")
        )

        assert stream.exists() == False
        assert stream.refresh_metadata.call_count == 1

    def test_exists_passes_other_errors(self):
        """
        Assert exists does not keep non 404 errors trapped
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)
        stream.refresh_metadata = Mock(side_effect=ValueError())

        with pytest.raises(ValueError):
            stream.exists()
        assert stream.refresh_metadata.call_count == 1

    ##########################################################################
    ## tag/annotation tests
    ##########################################################################

    def test_tags_returns_copy(self):
        """
        Assert tags returns a copy of the tags dict
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu, tags=token)

        assert stream.tags() is not token
        assert stream.tags() == token

    def test_tags_returns_cached_values(self):
        """
        Assert tags returns a copy of the tags dict
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(
            btrdb=BTrDB(Mock(Endpoint)), uuid=uu, tags=token, property_version=42
        )
        stream.refresh_metadata = Mock()

        assert stream.tags(refresh=False) == token
        stream.refresh_metadata.assert_not_called()

    def test_tags_forces_refresh_if_requested(self):
        """
        Assert tags calls refresh_metadata if requested even though a
        cached copy is available
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu, tags=token)
        stream.refresh_metadata = Mock()

        stream.tags(refresh=True)
        assert stream.refresh_metadata.call_count == 1

    def test_annotations_returns_copy_of_value(self):
        """
        Assert annotations returns a copy of the annotations dict
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(
            btrdb=BTrDB(Mock(Endpoint)), uuid=uu, annotations=token, property_version=42
        )
        stream.refresh_metadata = Mock()

        assert stream.annotations(refresh=False)[0] == token
        assert stream.annotations(refresh=False)[0] is not token
        assert stream.annotations(refresh=False)[1] == 42

    def test_annotations_returns_cached_values(self):
        """
        Assert annotations returns a copy of the annotations dict
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(
            btrdb=BTrDB(Mock(Endpoint)), uuid=uu, annotations=token, property_version=42
        )
        stream.refresh_metadata = Mock()

        assert stream.annotations(refresh=False)[0] == token
        stream.refresh_metadata.assert_not_called()

    def test_annotations_forces_refresh_if_requested(self):
        """
        Assert annotations calls refresh_metadata if requested even though a
        cached copy is available
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        token = {"cat": "dog"}
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu, annotations=token)
        stream.refresh_metadata = Mock()

        stream.annotations(refresh=True)
        assert stream.refresh_metadata.call_count == 1

    ##########################################################################
    ## windowing tests
    ##########################################################################

    def test_windows(self):
        """
        Assert windows returns tuples of data from Endpoint.windows
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        windows = [
            [
                (
                    StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
                    StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),
                ),
                42,
            ],
            [
                (
                    StatPointProto(time=3, min=4, mean=5, max=6, count=7, stddev=8),
                    StatPointProto(time=4, min=5, mean=6, max=7, count=8, stddev=9),
                ),
                42,
            ],
        ]
        expected = (
            (StatPoint(time=1, minv=2.0, meanv=3.0, maxv=4.0, count=5, stddev=6.0), 42),
            (StatPoint(time=2, minv=3.0, meanv=4.0, maxv=5.0, count=6, stddev=7.0), 42),
            (StatPoint(time=3, minv=4.0, meanv=5.0, maxv=6.0, count=7, stddev=8.0), 42),
            (StatPoint(time=4, minv=5.0, meanv=6.0, maxv=7.0, count=8, stddev=9.0), 42),
        )
        endpoint.windows = Mock(return_value=windows)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        params = {"start": 100, "end": 500, "width": 2}

        result = stream.windows(**params)
        assert result == expected
        assert isinstance(result, tuple)
        assert isinstance(result[0], tuple)
        stream._btrdb.ep.windows.assert_called_once_with(uu, 100, 500, 2, 0, 0)

    def test_aligned_windows(self):
        """
        Assert windows returns tuples of data from Endpoint.alignedWindows
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        windows = [
            [
                (
                    StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
                    StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),
                ),
                42,
            ],
            [
                (
                    StatPointProto(time=3, min=4, mean=5, max=6, count=7, stddev=8),
                    StatPointProto(time=4, min=5, mean=6, max=7, count=8, stddev=9),
                ),
                42,
            ],
        ]
        expected = (
            (StatPoint(time=1, minv=2.0, meanv=3.0, maxv=4.0, count=5, stddev=6.0), 42),
            (StatPoint(time=2, minv=3.0, meanv=4.0, maxv=5.0, count=6, stddev=7.0), 42),
            (StatPoint(time=3, minv=4.0, meanv=5.0, maxv=6.0, count=7, stddev=8.0), 42),
            (StatPoint(time=4, minv=5.0, meanv=6.0, maxv=7.0, count=8, stddev=9.0), 42),
        )
        endpoint.alignedWindows = Mock(return_value=windows)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        params = {"start": 100, "end": 500, "pointwidth": 1}

        result = stream.aligned_windows(**params)
        assert result == expected
        assert isinstance(result, tuple)
        assert isinstance(result[0], tuple)
        stream._btrdb.ep.alignedWindows.assert_called_once_with(uu, 100, 500, 1, 0)

    def test_count(self):
        """
        Test that stream count method uses aligned windows
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        windows = [
            [
                (
                    StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
                    StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),
                ),
                42,
            ],
            [
                (
                    StatPointProto(time=3, min=4, mean=5, max=6, count=7, stddev=8),
                    StatPointProto(time=4, min=5, mean=6, max=7, count=8, stddev=9),
                ),
                42,
            ],
        ]
        endpoint.alignedWindows = Mock(return_value=windows)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        assert stream.count(precise=False, version=0) == 26
        stream._btrdb.ep.alignedWindows.assert_called_once_with(
            uu, MINIMUM_TIME, MAXIMUM_TIME, 60, 0
        )

        stream.count(10, 1000, 8, version=1200)
        stream._btrdb.ep.alignedWindows.assert_called_with(uu, 10, 1000, 8, 1200)

    ##########################################################################
    ## earliest/latest tests
    ##########################################################################

    def test_earliest(self):
        """
        Assert earliest calls Endpoint.nearest
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(return_value=(RawPointProto(time=100, value=1.0), 42))

        point, ver = stream.earliest()
        assert (point, ver) == (RawPoint(100, 1.0), 42)
        endpoint.nearest.assert_called_once_with(uu, MINIMUM_TIME, 0, False)

    def test_earliest_swallows_exception(self):
        """
        Assert earliest returns None when endpoint throws exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=NoSuchPoint("next point does not exist"))

        assert stream.earliest() is None
        endpoint.nearest.assert_called_once_with(uu, MINIMUM_TIME, 0, False)

    def test_earliest_passes_exception(self):
        """
        Assert earliest reraises non 401 exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=BTrDBError("empty"))

        with pytest.raises(BTrDBError) as exc:
            stream.earliest()
        assert exc.value.args[0] == "empty"
        endpoint.nearest.assert_called_once_with(uu, MINIMUM_TIME, 0, False)

    def test_latest(self):
        """
        Assert latest calls Endpoint.nearest
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(return_value=(RawPointProto(time=100, value=1.0), 42))
        point, ver = stream.latest()

        assert (point, ver) == (RawPoint(100, 1.0), 42)
        endpoint.nearest.assert_called_once_with(uu, MAXIMUM_TIME, 0, True)

    def test_latest_swallows_exception(self):
        """
        Assert latest returns None when endpoint throws exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=NoSuchPoint("empty"))

        assert stream.latest() is None
        endpoint.nearest.assert_called_once_with(uu, MAXIMUM_TIME, 0, True)

    def test_latest_passes_exception(self):
        """
        Assert latest reraises non 401 exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=BTrDBError("empty"))

        with pytest.raises(BTrDBError) as exc:
            stream.latest()
        assert exc.value.args[0] == "empty"
        endpoint.nearest.assert_called_once_with(uu, MAXIMUM_TIME, 0, True)

    @patch("btrdb.stream.currently_as_ns")
    def test_currently(self, mocked):
        """
        Assert currently calls Endpoint.nearest
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(return_value=(RawPointProto(time=100, value=1.0), 42))
        ns_fake_time = 1514808000000000000
        mocked.return_value = ns_fake_time
        point, ver = stream.current()

        assert (point, ver) == (RawPoint(100, 1.0), 42)
        endpoint.nearest.assert_called_once_with(uu, ns_fake_time, 0, True)

    @patch("btrdb.stream.currently_as_ns")
    def test_currently_swallows_exception(self, mocked):
        """
        Assert currently returns None when endpoint throws exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=NoSuchPoint("empty"))
        ns_fake_time = 1514808000000000000
        mocked.return_value = ns_fake_time

        assert stream.current() is None
        endpoint.nearest.assert_called_once_with(uu, ns_fake_time, 0, True)

    @patch("btrdb.stream.currently_as_ns")
    def test_currently_passes_exception(self, mocked):
        """
        Assert currently reraises non 401 exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=BTrDBError("empty"))
        ns_fake_time = 1514808000000000000
        mocked.return_value = ns_fake_time

        with pytest.raises(BTrDBError) as exc:
            stream.current()
        assert exc.value.args[0] == "empty"
        endpoint.nearest.assert_called_once_with(uu, ns_fake_time, 0, True)

    ##########################################################################
    ## misc tests
    ##########################################################################

    def test_version(self):
        """
        Assert version calls and returns correct value from streamInfo
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=("", 0, {}, {}, 42))
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        assert stream.version() == 42
        stream._btrdb.ep.streamInfo.assert_called_once_with(uu, True, False)

    def test_insert(self):
        """
        Assert insert batches data to endpoint insert and returns version
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.insert = Mock(side_effect=[1, 2, 3])
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        data = list(zip(range(10000, 120000), map(float, range(110000))))
        version = stream.insert(data)

        assert (
            stream._btrdb.ep.insert.call_args_list[0][0][1] == data[:INSERT_BATCH_SIZE]
        )
        assert (
            stream._btrdb.ep.insert.call_args_list[1][0][1]
            == data[INSERT_BATCH_SIZE : 2 * INSERT_BATCH_SIZE]
        )
        assert (
            stream._btrdb.ep.insert.call_args_list[2][0][1]
            == data[2 * INSERT_BATCH_SIZE :]
        )
        assert version == 3

    def test_nearest(self):
        """
        Assert nearest calls Endpoint.nearest with correct arguments
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.nearest = Mock(return_value=(RawPointProto(time=100, value=1.0), 42))
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        point, version = stream.nearest(5, 10, False)
        stream._btrdb.ep.nearest.assert_called_once_with(uu, 5, 10, False)
        assert point == RawPoint(100, 1.0)
        assert version == 42

    def test_nearest_swallows_exception(self):
        """
        Assert nearest returns None when endpoint throws 401 exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=NoSuchPoint("next point does not exist"))

        assert stream.nearest(0, 0, False) is None
        endpoint.nearest.assert_called_once_with(uu, 0, 0, False)

    def test_nearest_passes_exception(self):
        """
        Assert nearest reraises non 401 exception
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)
        endpoint.nearest = Mock(side_effect=BTrDBError("foo"))

        with pytest.raises(BTrDBError) as exc:
            stream.nearest(0, 0, False)
        assert exc.value.args[0] == "foo"
        endpoint.nearest.assert_called_once_with(uu, 0, 0, False)

    def test_delete_range(self):
        """
        Assert delete_range calls Endpoint.deleteRange with correct arguments
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)

        stream.delete(5, 10)
        stream._btrdb.ep.deleteRange.assert_called_once_with(uu, 5, 10)

    def test_flush(self):
        """
        Assert flush calls Endpoint.flush with UUID
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)

        stream.flush()
        stream._btrdb.ep.flush.assert_called_once_with(uu)

    def test_obliterate(self):
        """
        Assert obliterate calls Endpoint.obliterate with UUID
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        stream = Stream(btrdb=BTrDB(Mock(Endpoint)), uuid=uu)

        stream.obliterate()
        stream._btrdb.ep.obliterate.assert_called_once_with(uu)

    def test_obliterate_allows_error(self):
        """
        Assert obliterate raises error if stream not found. (does not swallow error)
        """
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        endpoint = Mock(Endpoint)
        endpoint.obliterate = Mock(side_effect=StreamNotFoundError())
        stream = Stream(btrdb=BTrDB(endpoint), uuid=uu)

        with pytest.raises(BTrDBError):
            stream.obliterate()
        stream._btrdb.ep.obliterate.assert_called_once_with(uu)


##########################################################################
## StreamSet Tests
##########################################################################


class TestStreamSet(object):
    def test_create(self):
        """
        Assert we can create the object
        """
        StreamSet([Stream(None, "1"), Stream(None, "2"), Stream(None, "3")])

    @pytest.mark.parametrize(
        "empty_container", [(tuple()), (dict()), (list()), (set())]
    )
    def test_create_error_with_no_objects(self, empty_container):
        with pytest.raises(
            ValueError, match="Trying to create streamset with an empty list of streams"
        ):
            StreamSet(empty_container)

    ##########################################################################
    ## builtin / magic methods tests
    ##########################################################################

    def test_repr(self):
        """
        Assert StreamSet instance repr output
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        expected = "<StreamSet(4 streams)>"
        assert streams.__repr__() == expected

        data = [Stream(None, "1")]
        streams = StreamSet(data)
        expected = "<StreamSet(1 stream)>"
        assert streams.__repr__() == expected

    def test_str(self):
        """
        Assert StreamSet instance str output
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        expected = "StreamSet with 4 streams"
        assert str(streams) == expected

        data = [Stream(None, "1")]
        streams = StreamSet(data)
        expected = "StreamSet with 1 stream"
        assert str(streams) == expected

    def test_subscriptable(self):
        """
        Assert StreamSet instance is subscriptable
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        for index, val in enumerate(data):
            assert streams[index] == val

    def test_len(self):
        """
        Assert StreamSet instance support len
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        assert len(streams) == len(data)

    def test_iter(self):
        """
        Assert StreamSet instance support iteration
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        for index, stream in enumerate(streams):
            assert data[index] == stream

    def test_indexing(self):
        """
        Assert StreamSet instance supports indexing
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)

        # verify index lookup
        assert streams[-1] == data[-1]
        for idx in range(len(streams)):
            assert data[idx] == streams[idx]

        # verify slicing works
        assert streams[:2] == data[:2]

    def test_mapping(self):
        """
        Assert StreamSet instance support key mapping
        """
        uuids = [uuid.uuid4() for _ in range(4)]
        data = [Stream(None, uu) for uu in uuids]
        streams = StreamSet(data)

        # verify lookup with UUID
        for uu in uuids:
            assert streams[uu].uuid == uu

        # verify lookup with str
        for uu in uuids:
            assert streams[str(uu)].uuid == uu

        # verify raises KeyError
        missing = uuid.uuid4()
        with pytest.raises(KeyError) as e:
            streams[missing]
        assert str(missing) in str(e)

    def test_contains(self, stream1, stream2, stream3):
        """
        Assert StreamSet instance supports contains
        """
        data = [stream1, stream2, stream3]
        streams = StreamSet(data)
        assert "0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a" in streams

    def test_reverse(self):
        """
        Assert StreamSet instance supports reversal
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        assert list(reversed(streams)) == list(reversed(data))

    def test_to_list(self):
        """
        Assert StreamSet instance cast to list
        """
        data = [
            Stream(None, "11"),
            Stream(None, "22"),
            Stream(None, "dog"),
            Stream(None, "cat"),
        ]
        streams = StreamSet(data)
        assert list(streams) == data

    ##########################################################################
    ## allow_window tests
    ##########################################################################

    def test_allow_window(self, stream1, stream2, stream3):
        """
        Assert allow_window returns False if window already requested
        """
        streams = StreamSet([stream1, stream2, stream3])
        assert streams.allow_window == True

        streams.windows(30)
        assert streams.allow_window == False

        streams = StreamSet([stream1, stream2, stream3])
        streams.aligned_windows(30)
        assert streams.allow_window == False

    ##########################################################################
    ## _latest_versions tests
    ##########################################################################

    def test_latest_versions(self, stream1, stream2):
        """
        Assert _latest_versions returns correct values
        """
        streams = StreamSet([stream1, stream2])
        expected = {stream1.uuid: stream1.version(), stream2.uuid: stream2.version()}
        assert streams._latest_versions() == expected

    ##########################################################################
    ## pin_versions tests
    ##########################################################################

    def test_pin_versions_returns_self(self, stream1):
        """
        Assert pin_versions returns self
        """
        streams = StreamSet([stream1])
        expected = {stream1.uuid(): stream1.version()}
        result = streams.pin_versions(expected)
        assert streams is result

    def test_pin_versions_with_argument(self, stream1, stream2):
        """
        Assert pin_versions uses supplied version numbers
        """
        streams = StreamSet([stream1, stream2])
        expected = [3, 4]
        assert streams.pin_versions(expected) == streams
        assert streams._pinned_versions == expected

    def test_pin_versions_with_argument(self, stream1, stream2):
        """
        Assert pin_versions uses supplied version numbers
        """
        streams = StreamSet([stream1, stream2])
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        expected = {uu: 42}
        assert streams.pin_versions(expected) == streams
        assert streams._pinned_versions == expected

    def test_pin_versions_no_argument(self, stream1, stream2):
        """
        Assert pin_versions uses latest version numbers
        """
        streams = StreamSet([stream1, stream2])
        uu = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        expected = {uu: 42}

        assert streams.pin_versions(expected) == streams
        assert streams._pinned_versions == expected

    def test_pin_versions_raise_on_non_dict(self, stream1):
        """
        Assert pin_versions raises if versions argument is not dict
        """
        streams = StreamSet([stream1])
        expected = "INVALID DATA"

        with pytest.raises(TypeError) as e:
            streams.pin_versions(expected) == streams
        assert "dict" in str(e).lower()

    def test_pin_versions_raise_on_non_uuid_key(self):
        """
        Assert pin_versions raises if versions argument dict does not have UUID
        for keys
        """
        streams = StreamSet([Stream(None, "1")])
        expected = {"uuid": 42}

        with pytest.raises(TypeError) as e:
            streams.pin_versions(expected) == streams
        assert "uuid" in str(e).lower()

    ##########################################################################
    ## versions tests
    ##########################################################################

    def test_versions_no_pin(self, stream1, stream2):
        """
        Assert versions returns correctly if pin_versions not called
        """
        streams = StreamSet([stream1, stream2])
        expected = {stream1.uuid: stream1.version(), stream2.uuid: stream2.version()}

        assert streams.versions() == expected

    def test_versions_with_pin(self, stream1, stream2):
        """
        Assert versions returns correctly if pin_versions called
        """
        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        s1 = Mock(Stream)
        s1.version = Mock(return_value=11)
        s1.uuid = Mock(return_value=uu1)

        uu2 = uuid.UUID("17dbe387-89ea-42b6-864b-f505cdb483f5")
        s2 = Mock(Stream)
        s2.version = Mock(return_value=22)
        s2.uuid = Mock(return_value=uu2)

        streams = StreamSet([stream1, stream2])
        expected = {stream1.uuid(): 88, stream2.uuid(): 99}
        streams.pin_versions(expected)
        assert streams.versions() == expected

    ##########################################################################
    ## earliest/latest tests
    ##########################################################################

    def test_earliest(self, stream1, stream2):
        """
        Assert earliest returns correct time code
        """
        streams = StreamSet([stream1, stream2])
        assert streams.earliest() == (
            RawPoint(time=10, value=1),
            RawPoint(time=20, value=1),
        )

    def test_latest(self, stream1, stream2):
        """
        Assert latest returns correct time code
        """
        streams = StreamSet([stream1, stream2])
        assert streams.latest() == (
            RawPoint(time=10, value=1),
            RawPoint(time=20, value=1),
        )

    @patch("btrdb.stream.currently_as_ns")
    def test_current(self, mocked, stream1, stream2):
        """
        Assert current calls nearest with the current time
        """
        mocked.return_value = 15
        streams = StreamSet([stream1, stream2])
        streams.current()
        stream1.nearest.assert_called_once_with(15, version=11, backward=True)
        stream2.nearest.assert_called_once_with(15, version=22, backward=True)

    @patch("btrdb.stream.currently_as_ns")
    def test_currently_out_of_range(self, mocked, stream1, stream2):
        """
        Assert currently raises an exception if it is not filtered
        """
        mocked.return_value = 15
        streams = StreamSet([stream1, stream2])

        with pytest.raises(
            ValueError, match="current time is not included in filtered stream range"
        ):
            streams.filter(start=20, end=30).current()

        with pytest.raises(
            ValueError, match="current time is not included in filtered stream range"
        ):
            streams.filter(start=0, end=10).current()

    def test_count(self):
        """
        Test the stream set count method
        """
        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("4dadf38d-52a5-4b7a-ada9-a5d563f9538c")
        endpoint = Mock(Endpoint)
        endpoint.streamInfo = Mock(return_value=[-1, -1, -1, -1, 0])
        windows = [
            [
                (
                    StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
                    StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),
                ),
                42,
            ],
            [
                (
                    StatPointProto(time=3, min=4, mean=5, max=6, count=7, stddev=8),
                    StatPointProto(time=4, min=5, mean=6, max=7, count=8, stddev=9),
                ),
                42,
            ],
        ]
        endpoint.alignedWindows = Mock(return_value=windows)
        streams = StreamSet(
            [
                Stream(btrdb=BTrDB(endpoint), uuid=uu1),
                Stream(btrdb=BTrDB(endpoint), uuid=uu2),
            ]
        )

        assert streams.count() == 52
        endpoint.alignedWindows.assert_any_call(uu1, MINIMUM_TIME, MAXIMUM_TIME, 60, 0)
        endpoint.alignedWindows.assert_any_call(uu2, MINIMUM_TIME, MAXIMUM_TIME, 60, 0)

    def test_count_filtered(self):
        """
        Test the stream set count method with filters
        """
        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("4dadf38d-52a5-4b7a-ada9-a5d563f9538c")
        endpoint = Mock(Endpoint)
        endpoint.alignedWindows = Mock(return_value=[])
        streams = StreamSet(
            [
                Stream(btrdb=BTrDB(endpoint), uuid=uu1),
                Stream(btrdb=BTrDB(endpoint), uuid=uu2),
            ]
        )
        windows = [
            [
                (
                    StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),
                    StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),
                ),
                42,
            ],
            [
                (
                    StatPointProto(time=3, min=4, mean=5, max=6, count=7, stddev=8),
                    StatPointProto(time=4, min=5, mean=6, max=7, count=8, stddev=9),
                ),
                42,
            ],
        ]
        endpoint.alignedWindows = Mock(return_value=windows)

        streams = streams.filter(start=10, end=1000)
        streams.pin_versions({uu1: 42, uu2: 99})

        streams.count()
        endpoint.alignedWindows.assert_any_call(uu1, 10, 1000, 8, 42)
        endpoint.alignedWindows.assert_any_call(uu2, 10, 1000, 8, 99)

    ##########################################################################
    ## filter tests
    ##########################################################################

    def test_filter(self, stream1):
        """
        Assert filter creates and stores a StreamFilter object
        """
        streams = StreamSet([stream1])
        start, end = 1, 100
        streams = streams.filter(start=start, end=end)

        assert len(streams.filters) == 1
        assert streams.filters[0].start == start
        assert streams.filters[0].end == end
        assert isinstance(streams.filters[0], StreamFilter)

    def test_filter_returns_new_instance(self, stream1):
        """
        Assert filter returns new instance
        """
        streams = StreamSet([stream1])
        start, end = 1, 100
        other = streams.filter(start=start, end=end)

        assert other is not streams
        assert isinstance(other, streams.__class__)

    def test_filter_collection(self, stream1, stream2):
        """
        Assert filter collection works as intended
        """
        streams = StreamSet([stream1, stream2])

        # string arguments
        other = streams.filter(collection="fruits")
        assert other._streams == []
        other = streams.filter(collection="fruits/apple")
        assert other._streams == [stream1]
        other = streams.filter(collection="FRUITS/APPLE")
        assert other._streams == [stream1]
        other = streams.filter(collection="fruits*")
        assert other._streams == []

        # regex arguments
        other = streams.filter(collection=re.compile("fruits"))
        assert other._streams == [stream1, stream2]
        other = streams.filter(collection=re.compile("fruits.*"))
        assert other._streams == [stream1, stream2]

        type(stream1).collection = PropertyMock(return_value="foo/region-north")
        other = streams.filter(collection=re.compile("region-"))
        assert other._streams == [stream1]
        other = streams.filter(collection=re.compile("^region-"))
        assert other._streams == []
        other = streams.filter(collection=re.compile("foo/"))
        assert other._streams == [stream1]
        other = streams.filter(collection=re.compile("foo/z"))
        assert other._streams == []

        type(stream1).collection = PropertyMock(return_value="region.north/foo")
        other = streams.filter(collection=re.compile(r"region\."))
        assert other._streams == [stream1]

    def test_filter_name(self, stream1, stream2):
        """
        Assert filter name works as intended
        """
        streams = StreamSet([stream1, stream2])

        # string arguments
        other = streams.filter(name="blood")
        assert other._streams == [stream2]
        other = streams.filter(name="BLOOD")
        assert other._streams == [stream2]
        other = streams.filter(name="not_found")
        assert other._streams == []

        # regex arguments
        other = streams.filter(name=re.compile("blood"))
        assert other._streams == [stream2]
        other = streams.filter(name=re.compile("^blood$"))
        assert other._streams == [stream2]
        other = streams.filter(name=re.compile("oo"))
        assert other._streams == [stream2]
        other = streams.filter(name=re.compile("not_found"))
        assert other._streams == []

        type(stream1).name = PropertyMock(return_value="region-north")
        other = streams.filter(name=re.compile("region-"))
        assert other._streams == [stream1]
        other = streams.filter(name=re.compile(r"region\."))
        assert other._streams == []

        type(stream1).name = PropertyMock(return_value="region.north")
        other = streams.filter(name=re.compile(r"region\."))
        assert other._streams == [stream1]

    def test_filter_unit(self, stream1, stream2):
        """
        Assert filter unit works as intended
        """
        streams = StreamSet([stream1, stream2])

        # string arguments
        other = streams.filter(unit="volts")
        assert other._streams == [stream1]
        other = streams.filter(unit="VOLTS")
        assert other._streams == [stream1]
        other = streams.filter(unit="not_found")
        assert other._streams == []

        # regex arguments
        other = streams.filter(unit=re.compile("volts|amps"))
        assert other._streams == [stream1, stream2]
        other = streams.filter(unit=re.compile("volts"))
        assert other._streams == [stream1]
        other = streams.filter(unit=re.compile("v"))
        assert other._streams == [stream1]
        other = streams.filter(unit=re.compile("meters"))
        assert other._streams == []

    def test_filter_tags(self, stream1, stream2):
        """
        Assert filter annotations works as intended
        """
        streams = StreamSet([stream1, stream2])

        other = streams.filter(tags={"unit": "meters"})
        assert other._streams == []

        other = streams.filter(tags={"unit": "volts"})
        assert other._streams == [stream1]

        stream2.tags.return_value = {"name": "blood", "unit": "volts"}
        other = streams.filter(tags={"name": "blood", "unit": "volts"})
        assert other._streams == [stream2]
        other = streams.filter(tags={"unit": "volts"})
        assert other._streams == [stream1, stream2]

    def test_filter_annotations(self, stream1, stream2):
        """
        Assert filter annotations works as intended
        """
        streams = StreamSet([stream1, stream2])

        other = streams.filter(annotations={"owner": ""})
        assert other._streams == []

        other = streams.filter(annotations={"owner": "ABC"})
        assert other._streams == [stream1, stream2]

        other = streams.filter(annotations={"color": "red"})
        assert other._streams == [stream1]

        other = streams.filter(annotations={"owner": "ABC", "color": "red"})
        assert other._streams == [stream1]

    ##########################################################################
    ## clone tests
    ##########################################################################

    def test_clone_returns_new_object(self, stream1, stream2):
        """
        Assert that clone returns a different object
        """
        streams = StreamSet([stream1, stream2])
        clone = streams.clone()

        assert id(clone) != id(streams)
        assert clone._streams is not streams._streams
        assert clone._streams == streams._streams

    ##########################################################################
    ## windows tests
    ##########################################################################

    def test_windows_raises_valueerror_and_warning(self, stream1):
        """
        Assert that raises ValueError if arguments not castable to int
        """
        streams = StreamSet([stream1])
        with pytest.raises(ValueError) as exc:
            streams.windows("invalid", 42)
        assert "literal" in str(exc).lower()

        with pytest.warns(Warning, match="Overriding") as exc:
            streams.windows(42, "invalid")

    def test_windows_raises_if_not_allowed(self, stream1):
        """
        Assert windows raises Exception if not allowed
        """
        streams = StreamSet([stream1])
        streams.windows(8)

        with pytest.raises(Exception) as exc:
            streams.windows(10)
        assert "window operation is already requested" in str(exc).lower()

    def test_windows_returns_self(self, stream1):
        """
        Assert windows returns self
        """
        streams = StreamSet([stream1])
        result = streams.windows(10)
        assert result is streams

    def test_windows_stores_values(self, stream1):
        """
        Assert windows stores values
        """
        streams = StreamSet([stream1])
        result = streams.windows(10)

        # assert stores values
        assert result.width == 10
        assert result.depth == 0

    def test_windows_values_and_calls_to_endpoint(self):
        """
        assert windows result and endpoint calls are correct
        """
        endpoint = Mock(Endpoint)
        window1 = [
            [(StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),), 11]
        ]
        window2 = [
            [(StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),), 12]
        ]
        endpoint.windows = Mock(side_effect=[window1, window2])

        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("5d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        s1 = Stream(btrdb=BTrDB(endpoint), uuid=uu1)
        s2 = Stream(btrdb=BTrDB(endpoint), uuid=uu2)
        versions = {uu1: 11, uu2: 12}

        start, end, width = 1, 100, 1000
        streams = StreamSet([s1, s2])
        streams.pin_versions(versions)
        values = (
            streams.filter(start=start, end=end).windows(width=width, depth=0).values()
        )

        # assert endpoint calls have correct arguments, version
        expected = [
            call(uu1, start, end, width, 0, versions[uu1]),
            call(uu2, start, end, width, 0, versions[uu2]),
        ]
        assert endpoint.windows.call_args_list == expected

        # assert expected output
        expected = [
            [StatPoint(1, 2.0, 3.0, 4.0, 5, 6.0)],
            [StatPoint(2, 3.0, 4.0, 5.0, 6, 7.0)],
        ]
        assert values == expected

    def test_windows_rows_and_calls_to_endpoint(self):
        """
        assert windows rows result and endpoint calls are correct
        """
        endpoint = Mock(Endpoint)
        window1 = [
            [(StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),), 11]
        ]
        window2 = [
            [(StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),), 12]
        ]
        endpoint.windows = Mock(side_effect=[window1, window2])

        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("5d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        s1 = Stream(btrdb=BTrDB(endpoint), uuid=uu1)
        s2 = Stream(btrdb=BTrDB(endpoint), uuid=uu2)
        versions = {uu1: 11, uu2: 12}

        start, end, width = 1, 100, 1000
        streams = StreamSet([s1, s2])
        streams.pin_versions(versions)
        rows = streams.filter(start=start, end=end).windows(width=width, depth=0).rows()

        # assert endpoint calls have correct arguments, version
        expected = [
            call(uu1, start, end, width, 0, versions[uu1]),
            call(uu2, start, end, width, 0, versions[uu2]),
        ]
        assert endpoint.windows.call_args_list == expected

        # assert expected output
        expected = [
            (StatPoint(1, 2.0, 3.0, 4.0, 5, 6.0), None),
            (None, StatPoint(2, 3.0, 4.0, 5.0, 6, 7.0)),
        ]
        assert rows == expected

    ##########################################################################
    ## aligned_windows tests
    ##########################################################################

    def test_aligned_windows_raises_valueerror(self, stream1):
        """
        Assert aligned_windows raises ValueError if argument not castable to int
        """
        streams = StreamSet([stream1])
        with pytest.raises(ValueError) as exc:
            streams.aligned_windows("invalid")
        assert "literal" in str(exc).lower()

    def test_aligned_windows_raises_if_not_allowed(self, stream1):
        """
        Assert that aligned_windows raises Exception if not allowed
        """
        streams = StreamSet([stream1])
        streams.windows(8)

        with pytest.raises(Exception) as exc:
            streams.aligned_windows(20)
        assert "window operation is already requested" in str(exc).lower()

    def test_aligned_windows(self, stream1):
        """
        Assert aligned_windows stores objects
        """
        streams = StreamSet([stream1])
        result = streams.aligned_windows(20)
        assert streams.pointwidth == 20

    def test_aligned_windows_returns_self(self, stream1):
        """
        Assert aligned_windows returns self
        """
        streams = StreamSet([stream1])
        result = streams.aligned_windows(20)
        assert result is streams

    def test_aligned_windows_values_and_calls_to_endpoint(self):
        """
        assert aligned_windows result and endpoint calls are correct
        """
        endpoint = Mock(Endpoint)
        window1 = [
            [(StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),), 11]
        ]
        window2 = [
            [(StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),), 12]
        ]
        endpoint.alignedWindows = Mock(side_effect=[window1, window2])

        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("5d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        s1 = Stream(btrdb=BTrDB(endpoint), uuid=uu1)
        s2 = Stream(btrdb=BTrDB(endpoint), uuid=uu2)
        versions = {uu1: 11, uu2: 12}

        start, end, pointwidth = 1, 100, 25
        streams = StreamSet([s1, s2])
        streams.pin_versions(versions)
        values = (
            streams.filter(start=start, end=end)
            .aligned_windows(pointwidth=pointwidth)
            .values()
        )

        # assert endpoint calls have correct arguments, version
        expected = [
            call(uu1, start, end, pointwidth, versions[uu1]),
            call(uu2, start, end, pointwidth, versions[uu2]),
        ]
        assert endpoint.alignedWindows.call_args_list == expected

        # assert expected output
        expected = [
            [StatPoint(1, 2.0, 3.0, 4.0, 5, 6.0)],
            [StatPoint(2, 3.0, 4.0, 5.0, 6, 7.0)],
        ]
        assert values == expected

    def test_aligned_windows_rows_and_calls_to_endpoint(self):
        """
        assert aligned_windows rows result and endpoint calls are correct
        """
        endpoint = Mock(Endpoint)
        window1 = [
            [(StatPointProto(time=1, min=2, mean=3, max=4, count=5, stddev=6),), 11]
        ]
        window2 = [
            [(StatPointProto(time=2, min=3, mean=4, max=5, count=6, stddev=7),), 12]
        ]
        endpoint.alignedWindows = Mock(side_effect=[window1, window2])

        uu1 = uuid.UUID("0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        uu2 = uuid.UUID("5d22a53b-e2ef-4e0a-ab89-b2d48fb2592a")
        s1 = Stream(btrdb=BTrDB(endpoint), uuid=uu1)
        s2 = Stream(btrdb=BTrDB(endpoint), uuid=uu2)
        versions = {uu1: 11, uu2: 12}

        start, end, pointwidth = 1, 100, 25
        streams = StreamSet([s1, s2])
        streams.pin_versions(versions)
        rows = (
            streams.filter(start=start, end=end)
            .aligned_windows(pointwidth=pointwidth)
            .rows()
        )

        # assert endpoint calls have correct arguments, version
        expected = [
            call(uu1, start, end, pointwidth, versions[uu1]),
            call(uu2, start, end, pointwidth, versions[uu2]),
        ]
        assert endpoint.alignedWindows.call_args_list == expected

        # assert expected output
        expected = [
            (StatPoint(1, 2.0, 3.0, 4.0, 5, 6.0), None),
            (None, StatPoint(2, 3.0, 4.0, 5.0, 6, 7.0)),
        ]
        assert rows == expected

    ##########################################################################
    ## rows tests
    ##########################################################################

    def test_rows(self, stream1, stream2):
        """
        Assert rows returns correct values
        """
        stream1.values = Mock(
            return_value=iter(
                [
                    (RawPoint(time=1, value=1), 1),
                    (RawPoint(time=2, value=2), 1),
                    (RawPoint(time=3, value=3), 1),
                    (RawPoint(time=4, value=4), 1),
                ]
            )
        )
        stream2.values = Mock(
            return_value=iter(
                [(RawPoint(time=1, value=1), 2), (RawPoint(time=3, value=3), 2)]
            )
        )

        streams = StreamSet([stream1, stream2])
        rows = iter(streams.rows())

        assert next(rows) == (RawPoint(time=1, value=1), RawPoint(time=1, value=1))
        assert next(rows) == (RawPoint(time=2, value=2), None)
        assert next(rows) == (RawPoint(time=3, value=3), RawPoint(time=3, value=3))
        assert next(rows) == (RawPoint(time=4, value=4), None)

    ##########################################################################
    ## _params_from_filters tests
    ##########################################################################

    def test_params_from_filters_latest_value_wins(self, stream1):
        """
        Assert _params_from_filters returns latest setting
        """
        streams = StreamSet([stream1])
        assert streams.filters == []

        streams = streams.filter(start=1)
        assert streams._params_from_filters() == {"start": 1}

        streams = streams.filter(start=2)
        assert streams._params_from_filters() == {"start": 2}

    def test_params_from_filters_works(self, stream1):
        """
        Assert _params_from_filters returns correct values
        """
        streams = StreamSet([stream1])
        assert streams.filters == []

        streams = streams.filter(start=1, end=2)
        assert streams._params_from_filters() == {"start": 1, "end": 2}

        streams = streams.filter(start=9, end=10)
        assert streams._params_from_filters() == {"start": 9, "end": 10}

    ##########################################################################
    ## values tests
    ##########################################################################

    def test_values(self, stream1, stream2):
        """
        Assert values returns correct data
        """
        stream1_values = [
            (RawPoint(time=1, value=1), 1),
            (RawPoint(time=2, value=2), 1),
            (RawPoint(time=3, value=3), 1),
            (RawPoint(time=4, value=4), 1),
        ]
        stream1.values = Mock(return_value=iter(stream1_values))

        stream2_values = [
            (RawPoint(time=1, value=1), 2),
            (RawPoint(time=3, value=3), 2),
        ]
        stream2.values = Mock(return_value=iter(stream2_values))

        streams = StreamSet([stream1, stream2])
        assert streams.values() == [
            [t[0] for t in stream1_values],
            [t[0] for t in stream2_values],
        ]


##########################################################################
## StreamFilter Tests
##########################################################################


class TestStreamFilter(object):
    def test_create(self):
        """
        Assert we can create the object
        """
        StreamFilter(0, 1)

    def test_start_larger_or_equal(self):
        """
        Assert we raise ValueError if start is greater than/equal to end
        """
        with pytest.raises(ValueError):
            StreamFilter(1, 1)

    def test_start_valid(self):
        """
        Assert we can use any value castable to an int for start argument
        """
        sf = StreamFilter(start="100")
        assert sf.start == 100

        sf = StreamFilter(start=100.0)
        assert sf.start == 100

        sf = StreamFilter(start=100)
        assert sf.start == 100

    def test_start_invalid(self):
        """
        Assert we raise ValueError if start is greater than/equal to end
        """
        with pytest.raises(ValueError):
            StreamFilter(start="foo")



# btrdb.exceptions
# Module for custom exceptions
#
# Author:   PingThings
# Created:  Tue Dec 18 14:50:05 2018 -0500
#
# For license information, see LICENSE.txt
# ID: exceptions.py [] allen@pingthings.io $

##########################################################################
## Imports
##########################################################################

import inspect
import logging
import time
from functools import wraps

from grpc import RpcError

##########################################################################
## Module Variables and Constants
##########################################################################
MAX_RETRIES = 5
INITIAL_RETRY_DELAY = 3  # time to wait before first retry in seconds
RETRY_BACKOFF_FACTOR = (
    4  # exponential factor by which retry backoff increases between tries
)


##########################################################################
## Decorators
##########################################################################

# config logging for auto-retry
logging.basicConfig(
    format="%(asctime)s:%(levelname)s:%(message)s",
    level=logging.INFO,
    datefmt="%m/%d/%Y %I:%M:%S",
)


def retry(fn, *args, **kwargs):
    """
    decorates functions and retries it in the event of BTrDBErrors or RpcErrors
    """

    @wraps(fn)
    def retry_func(*args, **kwargs):
        # don't retry if arg is set to False
        if not kwargs.get("auto_retry"):
            return fn(*args, **kwargs)

        total_retries = kwargs.get("retries") or MAX_RETRIES
        retries = total_retries
        delay = kwargs.get("retry_delay") or INITIAL_RETRY_DELAY
        backoff = kwargs.get("retry_backoff") or RETRY_BACKOFF_FACTOR
        while retries > 0:
            try:
                return fn(*args, **kwargs)
            except (BTrDBError, RpcError) as e:
                msg = f"ERROR: {e}, attempting retry {total_retries-retries+1}/{total_retries} in {delay} seconds..."
                logging.info(msg)
                time.sleep(delay)
                retries -= 1
                delay *= backoff
                err = e
        handle_grpc_error(err)

    return retry_func


def consume_generator(fn, *args, **kwargs):
    # when a generator is passed back to the calling function, it may encounter an error
    # when trying to call next(), in that case we want to yield an Exception
    try:
        yield from fn(*args, **kwargs)
    except RpcError as e:
        handle_grpc_error(e)


def error_handler(fn):
    """
    decorates endpoint functions and checks for grpc.RpcErrors

    Parameters
    ----------
    fn: function
    """

    # allows input func to keep its name and metadata
    @wraps(fn)
    def wrap(*args, **kwargs):
        if inspect.isgeneratorfunction(fn):
            return consume_generator(fn, *args, **kwargs)
        try:
            return fn(*args, **kwargs)
        except RpcError as e:
            handle_grpc_error(e)

    return wrap


##########################################################################
## gRPC error handling
##########################################################################


# NOTE: this function relies on matching strings and isn't really ideal.
# this is more of a band-aid solution while we figure out how to send
# better errors from btrdb-server
def handle_grpc_error(err):
    """
    Called by endpoint functions when a gRPC error is encountered.
    Checks error details strings to catch known errors, if error is not
    known then a generic BTrDBError gets raised

    Parameters
    ----------
    err: Union[grpc.RpcError, btrdb.BTrDBError]
    """
    details = err.details()
    if isinstance(err, BTrDBError):
        if details == "[404] stream does not exist":
            raise StreamNotFoundError("Stream not found with provided uuid") from None
        elif details == "failed to connect to all addresses":
            raise ConnectionError("Failed to connect to BTrDB") from None
        elif any(str(e) in err.details() for e in BTRDB_SERVER_ERRORS):
            raise BTRDBServerError("An error has occured with btrdb-server") from None
        elif str(err.code()) == "StatusCode.PERMISSION_DENIED":
            raise PermissionDenied(details) from None
        raise BTrDBError(details) from None
    else:
        raise err


def check_proto_stat(stat):
    """
    Checks status of result after gRPC request and raises appropriate
    error based on status code

    Parameters
    ----------
    stat: btrdb_pb2.Status
    """
    code = stat.code
    if code != 0:
        if code in BTRDB_ERRORS:
            raise BTRDB_ERRORS[code](stat.msg)
        elif code in BTRDB_SERVER_ERRORS:
            raise BTRDBServerError(str(code) + ": " + stat.msg)
        raise BTrDBError(stat.msg)


##########################################################################
## BTrDB Exceptions
##########################################################################


class BTrDBError(Exception):
    """
    The primary exception for BTrDB errors.
    """

    pass


class ConnectionError(BTrDBError):
    """
    Raised when an error occurrs while trying to establish a connection with BTrDB.
    """

    pass


class StreamNotFoundError(BTrDBError):
    """
    Raised when attempting to perform an operation on a stream that does not exist in
    the specified BTrDB allocation.
    """

    pass


class CredentialsFileNotFound(FileNotFoundError, BTrDBError):
    """
    Raised when a credentials file could not be found.
    """

    pass


class ProfileNotFound(BTrDBError):
    """
    Raised when a requested profile could not be found in the credentials file.
    """

    pass


class BTRDBServerError(BTrDBError):
    """
    Raised when an error occurs with btrdb-server.
    """

    pass


class BTRDBTypeError(TypeError, BTrDBError):
    """
    Raised when attempting to perform an operation with an invalid type.
    """

    pass


class InvalidOperation(BTrDBError):
    """
    Raised when an invalid BTrDB operation has been requested.
    """

    pass


class StreamExists(InvalidOperation):
    """
    Raised when create() has been attempted and the uuid already exists.
    """

    pass


class AmbiguousStream(InvalidOperation):
    """
    Raised when create() has been attempted and uuid is different, but collection and tags already exist
    """

    pass


class PermissionDenied(InvalidOperation):
    """
    Raised when user does not have permission to perform an operation.
    """

    pass


class BTRDBValueError(ValueError, BTrDBError):
    """
    Raised when an invalid value has been passed to a BTrDB operation.
    """

    pass


class InvalidCollection(BTRDBValueError):
    """
    Raised when a collection name is invalid. It is either too long or not a valid string.
    """

    pass


class InvalidTagKey(BTRDBValueError):
    """
    Raised when a tag key is invalid. Must be one of ("name", "unit", "ingress", "distiller").
    """

    pass


class InvalidTagValue(BTRDBValueError):
    """
    Raised when a tag value is invalid. It is either too long or not a valid string.
    """

    pass


class InvalidTimeRange(BTRDBValueError):
    """
    Raised when insert data contains a timestamp outside the range of (btrdb.MINIMUM_TIME, btrdb.MAXIMUM_TIME)
    """

    pass


class InvalidPointWidth(BTRDBValueError):
    """
    Raised when attempting to use a pointwidth that is not a whole number between 0 and 64 (exclusive).
    """

    pass


class BadValue(BTRDBValueError):
    """
    Raised when attempting to insert data that contains non-float values such as None.
    """

    pass


class RecycledUUID(BTRDBValueError):
    """
    Raised when attempting to create a stream with a uuid that matches a previously deleted stream.
    """

    pass


class BadSQLValue(BTRDBValueError):
    """
    Raised when invalid parameters have been passed to metadata db.
    """

    pass


class VersionNotAvailable(BTRDBValueError):
    """
    Raised when querying a stream at a pruned or otherwise invalid version number.
    """

    pass


class NoSuchPoint(BTRDBValueError):
    """
    Raised when asking for next/previous point and there isn't one.
    """

    pass


##########################################################################
## Exception mapping
##########################################################################

# Errors that we have custom Exceptions for
BTRDB_ERRORS = {
    401: NoSuchPoint,
    404: StreamNotFoundError,
    407: InvalidCollection,
    408: InvalidTagKey,
    409: InvalidTagValue,
    413: InvalidTimeRange,
    415: InvalidPointWidth,
    417: StreamExists,
    418: AmbiguousStream,
    425: BadValue,
    429: RecycledUUID,
    441: BadSQLValue,
    450: VersionNotAvailable,
}

# All of these raise BTRDBServerError
BTRDB_SERVER_ERRORS = [
    402,  # ContextError
    403,  # InsertFailure
    405,  # WrongEndpoint
    414,  # InsertTooBig
    421,  # WrongArgs
    423,  # AnnotationVersionMismatch
    424,  # FaultInjectionDisabled
    426,  # ResourceDepleted
    427,  # InvalidVersions
    431,  # ObliterateDisabled
    432,  # CephError
    433,  # NodeExisted
    434,  # JournalError
    438,  # InvalidParameter
    440,  # MetadataConnectionError
    452,  # OverlappingTrimRange
    453,  # LockFailed
    454,  # NoLeafNode
    455,  # TierCacheError
    456,  # StreamEmpty
    457,  # TieredStorageBackendError
    458,  # TieredStorageOutOfBounds
    459,  # TieredStorageTemporaryError
    500,  # InvariantFailure
    501,  # NotImplemented
    502,  # UnsupportedRollback
]



# btrdb.point
# Module for Point classes
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: point.py [] allen@pingthings.io $

"""
Module for Point classes
"""

##########################################################################
## Imports
##########################################################################

from btrdb.grpcinterface import btrdb_pb2

##########################################################################
## Point Classes
##########################################################################


class RawPoint(object):
    """
    A point of data representing a single position within a time series. Each
    point contains a read-only time and value attribute.

    Parameters
    ----------
    time : int
        The time portion of a single value in the time series in nanoseconds
        since the Unix epoch.
    value : float
        The value of a time series at a single point in time.

    """

    __slots__ = ["_time", "_value"]

    def __init__(self, time, value):
        self._time = time
        self._value = value

    @property
    def time(self):
        """
        The time portion of a data point in nanoseconds since the Unix epoch.
        """
        return self._time

    @property
    def value(self):
        """
        The value portion of a data point as a float object.
        """
        return self._value

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.time, proto.value)

    @classmethod
    def from_proto_list(cls, proto_list):
        return [cls.from_proto(proto) for proto in proto_list]

    def __getitem__(self, index):
        if index == 0:
            return self.time
        elif index == 1:
            return self.value
        else:
            raise IndexError("RawPoint index out of range")

    @staticmethod
    def to_proto(point):
        return btrdb_pb2.RawPoint(time=point[0], value=point[1])

    @staticmethod
    def to_proto_list(points):
        return [RawPoint.to_proto(p) for p in points]

    def __repr__(self):
        return "RawPoint({0}, {1})".format(repr(self.time), repr(self.value))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not hasattr(other, "time") or not hasattr(other, "value"):
            return False
        return self.time == other.time and self.value == other.value


class StatPoint(object):
    """
    An aggregated data point representing a summary or rollup of one or more
    points of data within a single time series.

    This aggregation point provides for the min, mean, max, count, and standard
    deviation of all data values it spans.  It is returned by windowing queries
    such as `windows` or `aligned_windows`.

    Parameters
    ----------
    time : int
        The start time of the window which spans the aggregated values. Represented
        in nanoseconds since the Unix epoch.
    min : float
        The minimum value in a time series within a specified range of time.
    mean : float
        The mean value in a time series within a specified range of time.
    max : float
        The maximum value in a time series within a specified range of time.
    count : float
        The number of values in a time series within a specified range of time.
    stddev : float
        The standard deviation of values in a time series within a specified
        range of time.


    Notes
    -----
    This object may also be treated as a tuple by referencing the values
    according to position.

    .. code-block:: python

        // returns time
        val = point[0]

        // returns standard deviation
        val = point[5]


    """

    __slots__ = ["_time", "_min", "_mean", "_max", "_count", "_stddev"]

    def __init__(self, time, minv, meanv, maxv, count, stddev):
        self._time = time
        self._min = minv
        self._mean = meanv
        self._max = maxv
        self._count = count
        self._stddev = stddev

    @classmethod
    def from_proto(cls, proto):
        return cls(
            proto.time, proto.min, proto.mean, proto.max, proto.count, proto.stddev
        )

    @classmethod
    def from_proto_list(cls, proto_list):
        return [cls.from_proto(proto) for proto in proto_list]

    @property
    def time(self):
        """
        The starting time of the time series within the stat point
        """
        return self._time

    @property
    def min(self):
        """
        The minimum value of the time series within a range of time
        """
        return self._min

    @property
    def mean(self):
        """
        The mean value of the time series within a range of time
        """
        return self._mean

    @property
    def max(self):
        """
        The maximum value of the time series within a range of time
        """
        return self._max

    @property
    def count(self):
        """
        The number of values within the time series for a range of time
        """
        return self._count

    @property
    def stddev(self):
        """
        The standard deviation of the values of a time series within a range of time
        """
        return self._stddev

    def __getitem__(self, index):
        if index == 0:
            return self.time
        elif index == 1:
            return self.min
        elif index == 2:
            return self.mean
        elif index == 3:
            return self.max
        elif index == 4:
            return self.count
        elif index == 5:
            return self.stddev
        else:
            raise IndexError("RawPoint index out of range")

    def __repr__(self):
        return "StatPoint({}, {}, {}, {}, {}, {})".format(
            self.time, self.min, self.mean, self.max, self.count, self.stddev
        )

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        for attr in "time", "min", "mean", "max", "count", "stddev":
            if not hasattr(other, attr):
                return False

        return (
            self.time == other.time
            and self.min == other.min
            and self.mean == other.mean
            and self.max == other.max
            and self.count == other.count
            and self.stddev == other.stddev
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: btrdb.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x62trdb.proto\x12\x05v5api\"Q\n\x0fRawValuesParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\r\n\x05start\x18\x02 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x10\x12\x14\n\x0cversionMajor\x18\x04 \x01(\x04\"}\n\x11RawValuesResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12\x1f\n\x06values\x18\x04 \x03(\x0b\x32\x0f.v5api.RawPoint\"m\n\x14\x41rrowRawValuesParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\r\n\x05start\x18\x02 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x10\x12\x14\n\x0cversionMajor\x18\x04 \x01(\x04\x12\x15\n\rtemplateBytes\x18\x05 \x01(\x0c\"u\n\x16\x41rrowRawValuesResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12\x12\n\narrowBytes\x18\x04 \x01(\x0c\"\x85\x01\n\x16\x41rrowMultiValuesParams\x12\x0c\n\x04uuid\x18\x01 \x03(\x0c\x12\x14\n\x0cversionMajor\x18\x02 \x03(\x04\x12\r\n\x05start\x18\x03 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x04 \x01(\x10\x12\x14\n\x0csnapPeriodNs\x18\x05 \x01(\x03\x12\x15\n\rtemplateBytes\x18\x06 \x01(\x0c\"K\n\x18\x41rrowMultiValuesResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x12\n\narrowBytes\x18\x02 \x01(\x0c\"*\n\x0bRawPointVec\x12\x0c\n\x04time\x18\x01 \x01(\x10\x12\r\n\x05value\x18\x02 \x03(\x01\"j\n\x14\x41lignedWindowsParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\r\n\x05start\x18\x02 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x10\x12\x14\n\x0cversionMajor\x18\x04 \x01(\x04\x12\x12\n\npointWidth\x18\x05 \x01(\r\"\x83\x01\n\x16\x41lignedWindowsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12 \n\x06values\x18\x04 \x03(\x0b\x32\x10.v5api.StatPoint\"z\n\x1b\x41rrowAlignedWindowsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12\x12\n\narrowBytes\x18\x04 \x01(\x0c\"m\n\rWindowsParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\r\n\x05start\x18\x02 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x10\x12\x14\n\x0cversionMajor\x18\x04 \x01(\x04\x12\r\n\x05width\x18\x05 \x01(\x04\x12\r\n\x05\x64\x65pth\x18\x06 \x01(\r\"|\n\x0fWindowsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12 \n\x06values\x18\x04 \x03(\x0b\x32\x10.v5api.StatPoint\"s\n\x14\x41rrowWindowsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12\x12\n\narrowBytes\x18\x04 \x01(\x0c\"h\n\x10StreamInfoParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x13\n\x0bomitVersion\x18\x02 \x01(\x08\x12\x16\n\x0eomitDescriptor\x18\x03 \x01(\x08\x12\x19\n\x04role\x18\x64 \x01(\x0b\x32\x0b.v5api.Role\"\x8a\x01\n\x12StreamInfoResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12+\n\ndescriptor\x18\x04 \x01(\x0b\x32\x17.v5api.StreamDescriptor\"\x98\x01\n\x10StreamDescriptor\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x12\n\ncollection\x18\x02 \x01(\t\x12 \n\x04tags\x18\x03 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\'\n\x0b\x61nnotations\x18\x04 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\x17\n\x0fpropertyVersion\x18\x05 \x01(\x04\"\x82\x01\n\x1aSetStreamAnnotationsParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x1f\n\x17\x65xpectedPropertyVersion\x18\x02 \x01(\x04\x12#\n\x07\x63hanges\x18\x03 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\x10\n\x08removals\x18\x04 \x03(\t\";\n\x1cSetStreamAnnotationsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\"\x8a\x01\n\x13SetStreamTagsParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x1f\n\x17\x65xpectedPropertyVersion\x18\x02 \x01(\x04\x12 \n\x04tags\x18\x03 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\x12\n\ncollection\x18\x04 \x01(\t\x12\x0e\n\x06remove\x18\x05 \x03(\t\"4\n\x15SetStreamTagsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\"{\n\x0c\x43reateParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x12\n\ncollection\x18\x02 \x01(\t\x12 \n\x04tags\x18\x03 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\'\n\x0b\x61nnotations\x18\x04 \x03(\x0b\x32\x12.v5api.KeyOptValue\"-\n\x0e\x43reateResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\"@\n\x13MetadataUsageParams\x12\x0e\n\x06prefix\x18\x01 \x01(\t\x12\x19\n\x04role\x18\x64 \x01(\x0b\x32\x0b.v5api.Role\"y\n\x15MetadataUsageResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x1d\n\x04tags\x18\x02 \x03(\x0b\x32\x0f.v5api.KeyCount\x12$\n\x0b\x61nnotations\x18\x03 \x03(\x0b\x32\x0f.v5api.KeyCount\"&\n\x08KeyCount\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x04\"B\n\x15ListCollectionsParams\x12\x0e\n\x06prefix\x18\x01 \x01(\t\x12\x19\n\x04role\x18\x64 \x01(\x0b\x32\x0b.v5api.Role\"K\n\x17ListCollectionsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x13\n\x0b\x63ollections\x18\x02 \x03(\t\"\xab\x01\n\x13LookupStreamsParams\x12\x12\n\ncollection\x18\x01 \x01(\t\x12\x1a\n\x12isCollectionPrefix\x18\x02 \x01(\x08\x12 \n\x04tags\x18\x03 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\'\n\x0b\x61nnotations\x18\x04 \x03(\x0b\x32\x12.v5api.KeyOptValue\x12\x19\n\x04role\x18\x64 \x01(\x0b\x32\x0b.v5api.Role\"^\n\x15LookupStreamsResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12(\n\x07results\x18\x02 \x03(\x0b\x32\x17.v5api.StreamDescriptor\"S\n\rNearestParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04time\x18\x02 \x01(\x10\x12\x14\n\x0cversionMajor\x18\x03 \x01(\x04\x12\x10\n\x08\x62\x61\x63kward\x18\x04 \x01(\x08\"z\n\x0fNearestResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12\x1e\n\x05value\x18\x04 \x01(\x0b\x32\x0f.v5api.RawPoint\"U\n\rChangesParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x11\n\tfromMajor\x18\x02 \x01(\x04\x12\x0f\n\x07toMajor\x18\x03 \x01(\x04\x12\x12\n\nresolution\x18\x04 \x01(\r\"\x7f\n\x0f\x43hangesResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\x12#\n\x06ranges\x18\x04 \x03(\x0b\x32\x13.v5api.ChangedRange\"#\n\tRoundSpec\x12\x0e\n\x04\x62its\x18\x02 \x01(\x05H\x00\x42\x06\n\x04spec\"\x99\x01\n\x0cInsertParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04sync\x18\x02 \x01(\x08\x12(\n\x0cmerge_policy\x18\x04 \x01(\x0e\x32\x12.v5api.MergePolicy\x12\"\n\x08rounding\x18\x05 \x01(\x0b\x32\x10.v5api.RoundSpec\x12\x1f\n\x06values\x18\x03 \x03(\x0b\x32\x0f.v5api.RawPoint\"\x91\x01\n\x11\x41rrowInsertParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04sync\x18\x02 \x01(\x08\x12(\n\x0cmerge_policy\x18\x03 \x01(\x0e\x32\x12.v5api.MergePolicy\x12\"\n\x08rounding\x18\x04 \x01(\x0b\x32\x10.v5api.RoundSpec\x12\x12\n\narrowBytes\x18\x05 \x01(\x0c\"Y\n\x0eInsertResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\"8\n\x0c\x44\x65leteParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\r\n\x05start\x18\x02 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x10\"Y\n\x0e\x44\x65leteResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\"\x0c\n\nInfoParams\"\xa2\x01\n\x0cInfoResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x19\n\x04mash\x18\x02 \x01(\x0b\x32\x0b.v5api.Mash\x12\x14\n\x0cmajorVersion\x18\x03 \x01(\r\x12\x14\n\x0cminorVersion\x18\x04 \x01(\r\x12\r\n\x05\x62uild\x18\x05 \x01(\t\x12\x1f\n\x05proxy\x18\x06 \x01(\x0b\x32\x10.v5api.ProxyInfo\"#\n\tProxyInfo\x12\x16\n\x0eproxyEndpoints\x18\x01 \x03(\t\"1\n\x11\x46\x61ultInjectParams\x12\x0c\n\x04type\x18\x01 \x01(\x04\x12\x0e\n\x06params\x18\x02 \x01(\x0c\">\n\x13\x46\x61ultInjectResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\n\n\x02rv\x18\x02 \x01(\x0c\"\x1b\n\x0b\x46lushParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\"X\n\rFlushResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x14\n\x0cversionMajor\x18\x02 \x01(\x04\x12\x14\n\x0cversionMinor\x18\x03 \x01(\x04\" \n\x10ObliterateParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\"1\n\x12ObliterateResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\"\'\n\x08RawPoint\x12\x0c\n\x04time\x18\x01 \x01(\x10\x12\r\n\x05value\x18\x02 \x01(\x01\"`\n\tStatPoint\x12\x0c\n\x04time\x18\x01 \x01(\x10\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0c\n\x04mean\x18\x03 \x01(\x01\x12\x0b\n\x03max\x18\x04 \x01(\x01\x12\r\n\x05\x63ount\x18\x05 \x01(\x06\x12\x0e\n\x06stddev\x18\x06 \x01(\x01\"*\n\x0c\x43hangedRange\x12\r\n\x05start\x18\x01 \x01(\x10\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x10\">\n\x06Status\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x0b\n\x03msg\x18\x02 \x01(\t\x12\x19\n\x04mash\x18\x03 \x01(\x0b\x32\x0b.v5api.Mash\"\x98\x01\n\x04Mash\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12\x0e\n\x06leader\x18\x02 \x01(\t\x12\x16\n\x0eleaderRevision\x18\x03 \x01(\x03\x12\x13\n\x0btotalWeight\x18\x04 \x01(\x03\x12\x0f\n\x07healthy\x18\x05 \x01(\x08\x12\x10\n\x08unmapped\x18\x06 \x01(\x01\x12\x1e\n\x07members\x18\x07 \x03(\x0b\x32\r.v5api.Member\"\xc3\x01\n\x06Member\x12\x0c\n\x04hash\x18\x01 \x01(\r\x12\x10\n\x08nodename\x18\x02 \x01(\t\x12\n\n\x02up\x18\x03 \x01(\x08\x12\n\n\x02in\x18\x04 \x01(\x08\x12\x0f\n\x07\x65nabled\x18\x05 \x01(\x08\x12\r\n\x05start\x18\x06 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x07 \x01(\x03\x12\x0e\n\x06weight\x18\x08 \x01(\x03\x12\x16\n\x0ereadPreference\x18\t \x01(\x01\x12\x15\n\rhttpEndpoints\x18\n \x01(\t\x12\x15\n\rgrpcEndpoints\x18\x0b \x01(\t\"8\n\x0bKeyOptValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x03val\x18\x02 \x01(\x0b\x32\x0f.v5api.OptValue\"\x19\n\x08OptValue\x12\r\n\x05value\x18\x01 \x01(\t\"&\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"?\n\x0fStreamCSVConfig\x12\x0f\n\x07version\x18\x01 \x01(\x04\x12\r\n\x05label\x18\x02 \x01(\t\x12\x0c\n\x04uuid\x18\x03 \x01(\x0c\"\x9d\x02\n\x11GenerateCSVParams\x12\x35\n\tqueryType\x18\x01 \x01(\x0e\x32\".v5api.GenerateCSVParams.QueryType\x12\x11\n\tstartTime\x18\x02 \x01(\x03\x12\x0f\n\x07\x65ndTime\x18\x03 \x01(\x03\x12\x12\n\nwindowSize\x18\x04 \x01(\x04\x12\r\n\x05\x64\x65pth\x18\x05 \x01(\r\x12\x17\n\x0fincludeVersions\x18\x06 \x01(\x08\x12\'\n\x07streams\x18\x07 \x03(\x0b\x32\x16.v5api.StreamCSVConfig\"H\n\tQueryType\x12\x19\n\x15\x41LIGNED_WINDOWS_QUERY\x10\x00\x12\x11\n\rWINDOWS_QUERY\x10\x01\x12\r\n\tRAW_QUERY\x10\x02\"Q\n\x13GenerateCSVResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x10\n\x08isHeader\x18\x02 \x01(\x08\x12\x0b\n\x03row\x18\x03 \x03(\t\"J\n\x0eSQLQueryParams\x12\r\n\x05query\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x03(\t\x12\x19\n\x04role\x18\x64 \x01(\x0b\x32\x0b.v5api.Role\"D\n\x10SQLQueryResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x13\n\x0bSQLQueryRow\x18\x02 \x03(\x0c\"\x14\n\x04Role\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x94\x01\n\x19SetCompactionConfigParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x18\n\x10\x43ompactedVersion\x18\x02 \x01(\x04\x12>\n\x17reducedResolutionRanges\x18\x03 \x03(\x0b\x32\x1d.v5api.ReducedResolutionRange\x12\x0f\n\x07unused0\x18\x04 \x01(\x04\":\n\x1bSetCompactionConfigResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\")\n\x19GetCompactionConfigParams\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\"\xc1\x01\n\x1bGetCompactionConfigResponse\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x1a\n\x12LatestMajorVersion\x18\x02 \x01(\x04\x12\x18\n\x10\x43ompactedVersion\x18\x03 \x01(\x04\x12>\n\x17reducedResolutionRanges\x18\x04 \x03(\x0b\x32\x1d.v5api.ReducedResolutionRange\x12\x0f\n\x07unused0\x18\x05 \x01(\x04\"H\n\x16ReducedResolutionRange\x12\r\n\x05Start\x18\x01 \x01(\x03\x12\x0b\n\x03\x45nd\x18\x02 \x01(\x03\x12\x12\n\nResolution\x18\x03 \x01(\r\"K\n\x12SubscriptionUpdate\x12\'\n\x02op\x18\x01 \x01(\x0e\x32\x1b.v5api.SubscriptionUpdateOp\x12\x0c\n\x04uuid\x18\x02 \x03(\x0c\"Q\n\x10SubscriptionResp\x12\x1b\n\x04stat\x18\x01 \x01(\x0b\x32\r.v5api.Status\x12\x0c\n\x04uuid\x18\x02 \x01(\x0c\x12\x12\n\narrowBytes\x18\x03 \x01(\x0c*<\n\x0bMergePolicy\x12\t\n\x05NEVER\x10\x00\x12\t\n\x05\x45QUAL\x10\x01\x12\n\n\x06RETAIN\x10\x02\x12\x0b\n\x07REPLACE\x10\x03*7\n\x14SubscriptionUpdateOp\x12\r\n\tADD_UUIDS\x10\x00\x12\x10\n\x0cREMOVE_UUIDS\x10\x01\x32\xfd\r\n\x05\x42TrDB\x12?\n\tRawValues\x12\x16.v5api.RawValuesParams\x1a\x18.v5api.RawValuesResponse0\x01\x12N\n\x0e\x41rrowRawValues\x12\x1b.v5api.ArrowRawValuesParams\x1a\x1d.v5api.ArrowRawValuesResponse0\x01\x12T\n\x10\x41rrowMultiValues\x12\x1d.v5api.ArrowMultiValuesParams\x1a\x1f.v5api.ArrowMultiValuesResponse0\x01\x12N\n\x0e\x41lignedWindows\x12\x1b.v5api.AlignedWindowsParams\x1a\x1d.v5api.AlignedWindowsResponse0\x01\x12X\n\x13\x41rrowAlignedWindows\x12\x1b.v5api.AlignedWindowsParams\x1a\".v5api.ArrowAlignedWindowsResponse0\x01\x12\x39\n\x07Windows\x12\x14.v5api.WindowsParams\x1a\x16.v5api.WindowsResponse0\x01\x12\x43\n\x0c\x41rrowWindows\x12\x14.v5api.WindowsParams\x1a\x1b.v5api.ArrowWindowsResponse0\x01\x12@\n\nStreamInfo\x12\x17.v5api.StreamInfoParams\x1a\x19.v5api.StreamInfoResponse\x12^\n\x14SetStreamAnnotations\x12!.v5api.SetStreamAnnotationsParams\x1a#.v5api.SetStreamAnnotationsResponse\x12I\n\rSetStreamTags\x12\x1a.v5api.SetStreamTagsParams\x1a\x1c.v5api.SetStreamTagsResponse\x12\x34\n\x06\x43reate\x12\x13.v5api.CreateParams\x1a\x15.v5api.CreateResponse\x12Q\n\x0fListCollections\x12\x1c.v5api.ListCollectionsParams\x1a\x1e.v5api.ListCollectionsResponse0\x01\x12K\n\rLookupStreams\x12\x1a.v5api.LookupStreamsParams\x1a\x1c.v5api.LookupStreamsResponse0\x01\x12\x37\n\x07Nearest\x12\x14.v5api.NearestParams\x1a\x16.v5api.NearestResponse\x12\x39\n\x07\x43hanges\x12\x14.v5api.ChangesParams\x1a\x16.v5api.ChangesResponse0\x01\x12\x34\n\x06Insert\x12\x13.v5api.InsertParams\x1a\x15.v5api.InsertResponse\x12>\n\x0b\x41rrowInsert\x12\x18.v5api.ArrowInsertParams\x1a\x15.v5api.InsertResponse\x12\x34\n\x06\x44\x65lete\x12\x13.v5api.DeleteParams\x1a\x15.v5api.DeleteResponse\x12.\n\x04Info\x12\x11.v5api.InfoParams\x1a\x13.v5api.InfoResponse\x12\x43\n\x0b\x46\x61ultInject\x12\x18.v5api.FaultInjectParams\x1a\x1a.v5api.FaultInjectResponse\x12\x31\n\x05\x46lush\x12\x12.v5api.FlushParams\x1a\x14.v5api.FlushResponse\x12@\n\nObliterate\x12\x17.v5api.ObliterateParams\x1a\x19.v5api.ObliterateResponse\x12L\n\x10GetMetadataUsage\x12\x1a.v5api.MetadataUsageParams\x1a\x1c.v5api.MetadataUsageResponse\x12\x45\n\x0bGenerateCSV\x12\x18.v5api.GenerateCSVParams\x1a\x1a.v5api.GenerateCSVResponse0\x01\x12<\n\x08SQLQuery\x12\x15.v5api.SQLQueryParams\x1a\x17.v5api.SQLQueryResponse0\x01\x12\x43\n\tSubscribe\x12\x19.v5api.SubscriptionUpdate\x1a\x17.v5api.SubscriptionResp(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'btrdb_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_MERGEPOLICY']._serialized_start=6600
  _globals['_MERGEPOLICY']._serialized_end=6660
  _globals['_SUBSCRIPTIONUPDATEOP']._serialized_start=6662
  _globals['_SUBSCRIPTIONUPDATEOP']._serialized_end=6717
  _globals['_RAWVALUESPARAMS']._serialized_start=22
  _globals['_RAWVALUESPARAMS']._serialized_end=103
  _globals['_RAWVALUESRESPONSE']._serialized_start=105
  _globals['_RAWVALUESRESPONSE']._serialized_end=230
  _globals['_ARROWRAWVALUESPARAMS']._serialized_start=232
  _globals['_ARROWRAWVALUESPARAMS']._serialized_end=341
  _globals['_ARROWRAWVALUESRESPONSE']._serialized_start=343
  _globals['_ARROWRAWVALUESRESPONSE']._serialized_end=460
  _globals['_ARROWMULTIVALUESPARAMS']._serialized_start=463
  _globals['_ARROWMULTIVALUESPARAMS']._serialized_end=596
  _globals['_ARROWMULTIVALUESRESPONSE']._serialized_start=598
  _globals['_ARROWMULTIVALUESRESPONSE']._serialized_end=673
  _globals['_RAWPOINTVEC']._serialized_start=675
  _globals['_RAWPOINTVEC']._serialized_end=717
  _globals['_ALIGNEDWINDOWSPARAMS']._serialized_start=719
  _globals['_ALIGNEDWINDOWSPARAMS']._serialized_end=825
  _globals['_ALIGNEDWINDOWSRESPONSE']._serialized_start=828
  _globals['_ALIGNEDWINDOWSRESPONSE']._serialized_end=959
  _globals['_ARROWALIGNEDWINDOWSRESPONSE']._serialized_start=961
  _globals['_ARROWALIGNEDWINDOWSRESPONSE']._serialized_end=1083
  _globals['_WINDOWSPARAMS']._serialized_start=1085
  _globals['_WINDOWSPARAMS']._serialized_end=1194
  _globals['_WINDOWSRESPONSE']._serialized_start=1196
  _globals['_WINDOWSRESPONSE']._serialized_end=1320
  _globals['_ARROWWINDOWSRESPONSE']._serialized_start=1322
  _globals['_ARROWWINDOWSRESPONSE']._serialized_end=1437
  _globals['_STREAMINFOPARAMS']._serialized_start=1439
  _globals['_STREAMINFOPARAMS']._serialized_end=1543
  _globals['_STREAMINFORESPONSE']._serialized_start=1546
  _globals['_STREAMINFORESPONSE']._serialized_end=1684
  _globals['_STREAMDESCRIPTOR']._serialized_start=1687
  _globals['_STREAMDESCRIPTOR']._serialized_end=1839
  _globals['_SETSTREAMANNOTATIONSPARAMS']._serialized_start=1842
  _globals['_SETSTREAMANNOTATIONSPARAMS']._serialized_end=1972
  _globals['_SETSTREAMANNOTATIONSRESPONSE']._serialized_start=1974
  _globals['_SETSTREAMANNOTATIONSRESPONSE']._serialized_end=2033
  _globals['_SETSTREAMTAGSPARAMS']._serialized_start=2036
  _globals['_SETSTREAMTAGSPARAMS']._serialized_end=2174
  _globals['_SETSTREAMTAGSRESPONSE']._serialized_start=2176
  _globals['_SETSTREAMTAGSRESPONSE']._serialized_end=2228
  _globals['_CREATEPARAMS']._serialized_start=2230
  _globals['_CREATEPARAMS']._serialized_end=2353
  _globals['_CREATERESPONSE']._serialized_start=2355
  _globals['_CREATERESPONSE']._serialized_end=2400
  _globals['_METADATAUSAGEPARAMS']._serialized_start=2402
  _globals['_METADATAUSAGEPARAMS']._serialized_end=2466
  _globals['_METADATAUSAGERESPONSE']._serialized_start=2468
  _globals['_METADATAUSAGERESPONSE']._serialized_end=2589
  _globals['_KEYCOUNT']._serialized_start=2591
  _globals['_KEYCOUNT']._serialized_end=2629
  _globals['_LISTCOLLECTIONSPARAMS']._serialized_start=2631
  _globals['_LISTCOLLECTIONSPARAMS']._serialized_end=2697
  _globals['_LISTCOLLECTIONSRESPONSE']._serialized_start=2699
  _globals['_LISTCOLLECTIONSRESPONSE']._serialized_end=2774
  _globals['_LOOKUPSTREAMSPARAMS']._serialized_start=2777
  _globals['_LOOKUPSTREAMSPARAMS']._serialized_end=2948
  _globals['_LOOKUPSTREAMSRESPONSE']._serialized_start=2950
  _globals['_LOOKUPSTREAMSRESPONSE']._serialized_end=3044
  _globals['_NEARESTPARAMS']._serialized_start=3046
  _globals['_NEARESTPARAMS']._serialized_end=3129
  _globals['_NEARESTRESPONSE']._serialized_start=3131
  _globals['_NEARESTRESPONSE']._serialized_end=3253
  _globals['_CHANGESPARAMS']._serialized_start=3255
  _globals['_CHANGESPARAMS']._serialized_end=3340
  _globals['_CHANGESRESPONSE']._serialized_start=3342
  _globals['_CHANGESRESPONSE']._serialized_end=3469
  _globals['_ROUNDSPEC']._serialized_start=3471
  _globals['_ROUNDSPEC']._serialized_end=3506
  _globals['_INSERTPARAMS']._serialized_start=3509
  _globals['_INSERTPARAMS']._serialized_end=3662
  _globals['_ARROWINSERTPARAMS']._serialized_start=3665
  _globals['_ARROWINSERTPARAMS']._serialized_end=3810
  _globals['_INSERTRESPONSE']._serialized_start=3812
  _globals['_INSERTRESPONSE']._serialized_end=3901
  _globals['_DELETEPARAMS']._serialized_start=3903
  _globals['_DELETEPARAMS']._serialized_end=3959
  _globals['_DELETERESPONSE']._serialized_start=3961
  _globals['_DELETERESPONSE']._serialized_end=4050
  _globals['_INFOPARAMS']._serialized_start=4052
  _globals['_INFOPARAMS']._serialized_end=4064
  _globals['_INFORESPONSE']._serialized_start=4067
  _globals['_INFORESPONSE']._serialized_end=4229
  _globals['_PROXYINFO']._serialized_start=4231
  _globals['_PROXYINFO']._serialized_end=4266
  _globals['_FAULTINJECTPARAMS']._serialized_start=4268
  _globals['_FAULTINJECTPARAMS']._serialized_end=4317
  _globals['_FAULTINJECTRESPONSE']._serialized_start=4319
  _globals['_FAULTINJECTRESPONSE']._serialized_end=4381
  _globals['_FLUSHPARAMS']._serialized_start=4383
  _globals['_FLUSHPARAMS']._serialized_end=4410
  _globals['_FLUSHRESPONSE']._serialized_start=4412
  _globals['_FLUSHRESPONSE']._serialized_end=4500
  _globals['_OBLITERATEPARAMS']._serialized_start=4502
  _globals['_OBLITERATEPARAMS']._serialized_end=4534
  _globals['_OBLITERATERESPONSE']._serialized_start=4536
  _globals['_OBLITERATERESPONSE']._serialized_end=4585
  _globals['_RAWPOINT']._serialized_start=4587
  _globals['_RAWPOINT']._serialized_end=4626
  _globals['_STATPOINT']._serialized_start=4628
  _globals['_STATPOINT']._serialized_end=4724
  _globals['_CHANGEDRANGE']._serialized_start=4726
  _globals['_CHANGEDRANGE']._serialized_end=4768
  _globals['_STATUS']._serialized_start=4770
  _globals['_STATUS']._serialized_end=4832
  _globals['_MASH']._serialized_start=4835
  _globals['_MASH']._serialized_end=4987
  _globals['_MEMBER']._serialized_start=4990
  _globals['_MEMBER']._serialized_end=5185
  _globals['_KEYOPTVALUE']._serialized_start=5187
  _globals['_KEYOPTVALUE']._serialized_end=5243
  _globals['_OPTVALUE']._serialized_start=5245
  _globals['_OPTVALUE']._serialized_end=5270
  _globals['_KEYVALUE']._serialized_start=5272
  _globals['_KEYVALUE']._serialized_end=5310
  _globals['_STREAMCSVCONFIG']._serialized_start=5312
  _globals['_STREAMCSVCONFIG']._serialized_end=5375
  _globals['_GENERATECSVPARAMS']._serialized_start=5378
  _globals['_GENERATECSVPARAMS']._serialized_end=5663
  _globals['_GENERATECSVPARAMS_QUERYTYPE']._serialized_start=5591
  _globals['_GENERATECSVPARAMS_QUERYTYPE']._serialized_end=5663
  _globals['_GENERATECSVRESPONSE']._serialized_start=5665
  _globals['_GENERATECSVRESPONSE']._serialized_end=5746
  _globals['_SQLQUERYPARAMS']._serialized_start=5748
  _globals['_SQLQUERYPARAMS']._serialized_end=5822
  _globals['_SQLQUERYRESPONSE']._serialized_start=5824
  _globals['_SQLQUERYRESPONSE']._serialized_end=5892
  _globals['_ROLE']._serialized_start=5894
  _globals['_ROLE']._serialized_end=5914
  _globals['_SETCOMPACTIONCONFIGPARAMS']._serialized_start=5917
  _globals['_SETCOMPACTIONCONFIGPARAMS']._serialized_end=6065
  _globals['_SETCOMPACTIONCONFIGRESPONSE']._serialized_start=6067
  _globals['_SETCOMPACTIONCONFIGRESPONSE']._serialized_end=6125
  _globals['_GETCOMPACTIONCONFIGPARAMS']._serialized_start=6127
  _globals['_GETCOMPACTIONCONFIGPARAMS']._serialized_end=6168
  _globals['_GETCOMPACTIONCONFIGRESPONSE']._serialized_start=6171
  _globals['_GETCOMPACTIONCONFIGRESPONSE']._serialized_end=6364
  _globals['_REDUCEDRESOLUTIONRANGE']._serialized_start=6366
  _globals['_REDUCEDRESOLUTIONRANGE']._serialized_end=6438
  _globals['_SUBSCRIPTIONUPDATE']._serialized_start=6440
  _globals['_SUBSCRIPTIONUPDATE']._serialized_end=6515
  _globals['_SUBSCRIPTIONRESP']._serialized_start=6517
  _globals['_SUBSCRIPTIONRESP']._serialized_end=6598
  _globals['_BTRDB']._serialized_start=6720
  _globals['_BTRDB']._serialized_end=8509
# @@protoc_insertion_point(module_scope)



# tests.test_credentials
# Testing for the btrdb.utils.credentials module
#
# Author:   PingThings
# Created:  Thu May 09 14:29:21 2019 -0400
#
# For license information, see LICENSE.txt
# ID: test_credentials.py [] allen@pingthings.io $

"""
Testing for the btrdb.utils.credentials module
"""

##########################################################################
## Imports
##########################################################################

from unittest.mock import patch

import pytest

from btrdb.exceptions import *
from btrdb.utils.credentials import *

##########################################################################
## Tests
##########################################################################


class TestLoadCredentials(object):
    @patch("builtins.open")
    def test_raises_err_if_credentials_not_found(self, mock_open):
        """
        Assert CredentialsFileNotFound is raised if credentials.yaml is not found
        """
        mock_open.side_effect = FileNotFoundError("foo")
        with pytest.raises(CredentialsFileNotFound):
            load_credentials_from_file()


class TestLoadProfile(object):
    def setup_method(self):
        for env in ["BTRDB_ENDPOINTS", "BTRDB_PROFILE", "BTRDB_API_KEY"]:
            try:
                del os.environ[env]
            except KeyError:
                pass

    @patch("btrdb.utils.credentials.load_credentials_from_file")
    def test_raises_err_if_profile_not_found(self, mock_credentials):
        """
        Assert ProfileNotFound is raised if profile is asked for but not found
        """
        mock_credentials.return_value = {
            "duck": {
                "btrdb": {
                    "endpoints": "192.168.1.100:4410",
                    "api_key": "111222333",
                }
            }
        }
        with pytest.raises(ProfileNotFound):
            credentials_by_profile("horse")

    @patch("btrdb.utils.credentials.load_credentials_from_file")
    def test_returns_requested_profile(self, mock_credentials):
        """
        Assert returns correct profile
        """
        mock_credentials.return_value = {
            "default": {"btrdb": {"endpoints": "default"}},
            "duck": {"btrdb": {"endpoints": "duck"}},
        }
        assert credentials_by_profile("duck")["endpoints"] == "duck"

    @patch("btrdb.utils.credentials.load_credentials_from_file")
    def test_returns_default_profile(self, mock_credentials):
        """
        Assert default profile is returned
        """
        mock_credentials.return_value = {
            "duck": {
                "btrdb": {
                    "endpoints": "192.168.1.100:4411",
                    "api_key": "111222333",
                },
                "name": "duck",
            },
            "default": {
                "btrdb": {
                    "endpoints": "192.168.1.222:4411",
                    "api_key": "333222111",
                },
                "name": "default",
            },
        }
        assert credentials_by_profile()["apikey"] == "333222111"

    @patch("btrdb.utils.credentials.load_credentials_from_file")
    def test_returns_no_default_profile(self, mock_credentials):
        """
        Assert empty credentials are returned if no default profile
        """
        mock_credentials.return_value = {
            "duck": {
                "btrdb": {
                    "endpoints": "192.168.1.100:4411",
                    "api_key": "111222333",
                },
                "name": "duck",
            },
        }
        assert credentials_by_profile() == {}


class TestCredentials(object):
    def setup_method(self):
        for env in ["BTRDB_ENDPOINTS", "BTRDB_PROFILE", "BTRDB_API_KEY"]:
            try:
                del os.environ[env]
            except KeyError:
                pass

    @patch("btrdb.utils.credentials.load_credentials_from_file")
    @patch("os.path.exists")
    def test_checks_file_existence(self, mock_exists, mock_load):
        mock_exists.return_value = False
        credentials()
        assert not mock_load.called
